#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
# Copyright (c) 2025 DLR-SE Department of System Evolution and Operation
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import argparse
import pathlib

import pandas as pd

# import numpy as np
from seal5.logging import Logger


logger = Logger("examples." + __name__)


def main():
    """Main function."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser = argparse.ArgumentParser(description="Merge SEAL5 Test Results")
    parser.add_argument("status_file", type=str, help="Path to the status file")
    parser.add_argument("properties_file", type=str, help="Path to the file of instruction properties")
    parser.add_argument("test_results", type=str, help="Path to the file containing seal5 test result file name")
    parser.add_argument("test_cv_file", type=str, help="Path to the file containing seal5 test coverage results")
    args = parser.parse_args()

    if args.status_file:
        status_df = pd.read_csv(pathlib.Path(args.status_file))
    else:
        raise RuntimeError("Path to status file must be specified")

    if args.test_results:
        try:
            test_result_df = pd.read_csv(pathlib.Path(args.test_results))
            test_result_df = test_result_df[~pd.isna(test_result_df["instr"])]
        except pd.errors.EmptyDataError:
            logger.warning("Skipping empty file: %s", args.test_results)
            test_result_df = None
    else:
        raise RuntimeError("Path to Test Results file must be specified")

    if args.properties_file:
        props_df = pd.read_csv(pathlib.Path(args.properties_file))
    else:
        raise RuntimeError("Path to Instruction Properties file must be specified")

    if args.test_cv_file:
        test_cv_df = pd.read_csv(pathlib.Path(args.test_cv_file))
    else:
        raise RuntimeError("Path to Test Coverage file must be specified")

    status_props_df = pd.merge(status_df, props_df)
    if test_result_df is not None and len(test_result_df) > 0:
        stat_prop_result_df = pd.merge(
            status_props_df, test_result_df, on=["model", "set", "xlen", "instr"], how="left"
        )
    else:
        stat_prop_result_df = status_props_df

    stat_prop_result_df = stat_prop_result_df.sort_values("model")

    stat_prop_result_df.fillna(value=0, inplace=True)

    if test_cv_df is not None and len(test_cv_df) > 0:
        stat_prop_result_cv_df = pd.merge(
            stat_prop_result_df, test_cv_df, on=["model", "set", "xlen", "instr"], how="left"
        )
    else:
        stat_prop_result_cv_df = stat_prop_result_df

    def save_data_frames_as_html_to_file(dataframe, filename):
        with open(filename, "w", encoding="utf-8") as html_file:
            dataframe.to_html(buf=html_file, index=True)

    def rounded_mean(x):
        return x.mean().round(0).astype(int)

    stat_prop_result_cv_df.set_index(["model", "enc_format", "opcode"], inplace=True)

    stat_prop_result_all = stat_prop_result_cv_df.groupby(level=[0, 1, 2])[
        ["instr", "n_success", "n_failed", "n_skipped", "n_total", "status"]
    ].agg(
        {
            "instr": "unique",
            "status": "unique",
            "n_success": rounded_mean,
            "n_failed": "mean",
            "n_skipped": "mean",
            "n_total": "mean",
        }
    )

    if test_result_df is not None and len(test_result_df) > 0:
        stat_prop_result_cv_test = stat_prop_result_cv_df.groupby(level=[0, 1, 2])[
            ["n_pass", "n_fail", "n_tests", "state"]
        ].agg({"n_pass": rounded_mean, "n_fail": rounded_mean, "n_tests": rounded_mean, "state": "unique"})
    else:
        stat_prop_result_cv_test = None

    if test_cv_df is not None and len(test_cv_df) > 0:
        stat_prop_result_coverage = stat_prop_result_cv_df.groupby(level=[0, 1, 2])[
            ["n_exists", "n_required", "n_optional", "n_required_exists", "n_optional_exists", "coverage"]
        ].agg(
            {
                "n_exists": rounded_mean,
                "n_required": rounded_mean,
                "n_optional": rounded_mean,
                "n_required_exists": rounded_mean,
                "n_optional_exists": rounded_mean,
                "coverage": "unique",
            }
        )
    else:
        stat_prop_result_coverage = None

    def calc_stage_percentage():
        perc_success = (
            ((stat_prop_result_all["n_success"] * 100) / stat_prop_result_all["n_total"]).astype(int)
        ).astype(str)
        perc_skipped = (
            ((stat_prop_result_all["n_skipped"] * 100) / stat_prop_result_all["n_total"]).astype(int)
        ).astype(str)
        perc_failed = (((stat_prop_result_all["n_failed"] * 100) / stat_prop_result_all["n_total"]).astype(int)).astype(
            str
        )

        perc_result = "[" + perc_success + "%" + " / " + perc_skipped + "%" + " / " + perc_failed + "%" + "]"
        return perc_result

    # def calc_stage_percentage_sum():
    #     perc_success = (
    #         ((stat_prop_result_all["n_success"].sum() * 100) / stat_prop_result_all["n_total"].sum()).astype(int)
    #     ).astype(str)
    #     perc_skipped = (
    #         ((stat_prop_result_all["n_skipped"].sum() * 100) / stat_prop_result_all["n_total"].sum()).astype(int)
    #     ).astype(str)
    #     perc_failed = (
    #         ((stat_prop_result_all["n_failed"].sum() * 100) / stat_prop_result_all["n_total"].sum()).astype(int)
    #     ).astype(str)

    #     perc_result = "[" + perc_success + "%" + " / " + perc_skipped + "%" + " / " + perc_failed + "%" + "]"
    #     return perc_result

    def calc_test_percentage():
        perc_pass = (stat_prop_result_cv_test["n_pass"] * 100) / stat_prop_result_cv_test["n_tests"]
        perc_fail = (stat_prop_result_cv_test["n_fail"] * 100) / stat_prop_result_cv_test["n_tests"]
        #       if(np.isnan(perc_pass)):
        #         perc_pass = 0;
        perc_result = "[" + round(perc_pass, 0).astype(str) + "%" + " / " + round(perc_fail, 0).astype(str) + "%" + "]"
        return perc_result

    # def calc_test_percentage_sum():
    #     perc_pass = (
    #         (stat_prop_result_cv_test["n_pass"].sum() * 100) / stat_prop_result_cv_test["n_tests"].sum()
    #     ).astype(int)
    #     perc_fail = (
    #         (stat_prop_result_cv_test["n_fail"].sum() * 100) / stat_prop_result_cv_test["n_tests"].sum()
    #     ).astype(int)

    #     perc_result = "[" + round(perc_pass, 0).astype(str) + "%" + " / " + round(perc_fail, 0).astype(str)
    #         + "%" + "]"
    #     return perc_result

    def calc_coverage_percentage():
        perc_required = (
            (stat_prop_result_coverage["n_required_exists"] * 100) / stat_prop_result_coverage["n_required"]
        ).astype(
            int
        )  # .astype(str
        perc_optional = (
            (stat_prop_result_coverage["n_optional_exists"] * 100) / stat_prop_result_coverage["n_optional"]
        ).astype(
            int
        )  # .astype(str)
        perc_exists = (
            (stat_prop_result_coverage["n_exists"] * 100)
            / (stat_prop_result_coverage["n_required"] + stat_prop_result_coverage["n_optional"])
        ).astype(
            int
        )  # .astype(str)

        perc_result = (
            "["
            + round(perc_required, 0).astype(str)
            + "%"
            + " / "
            + round(perc_optional, 0).astype(str)
            + "%"
            + " / "
            + round(perc_exists, 0).astype(str)
            + "%]"
        )
        return perc_result

    stat_result_summary = (
        stat_prop_result_all["n_success"].astype(str)
        + " / "
        + stat_prop_result_all["n_skipped"].astype(str)
        + " / "
        + stat_prop_result_all["n_failed"].astype(str)
        + " "
        + calc_stage_percentage()
    )
    stat_prop_result_all.insert(loc=1, column="Status_Summary: (Passed/Skipped/Failed) %", value=stat_result_summary)

    if stat_prop_result_cv_test is not None:
        stat_test_summary = (
            stat_prop_result_cv_test["n_pass"].astype(str)
            + " / "
            + stat_prop_result_cv_test["n_fail"].astype(str)
            + " "
            + calc_test_percentage()
        )
        stat_prop_result_cv_test.insert(
            loc=1, column="Summary Test Results: (Passed/Failed) % ", value=stat_test_summary
        )

    if stat_prop_result_coverage is not None:
        stat_prop_coverage_summary = (
            stat_prop_result_coverage["n_required_exists"].astype(str)
            + " : "
            + stat_prop_result_coverage["n_required"].astype(str)
            + " / "
            + stat_prop_result_coverage["n_optional_exists"].astype(str)
            + " : "
            + stat_prop_result_coverage["n_optional"].astype(str)
            + " / "
            + stat_prop_result_coverage["n_exists"].astype(str)
            + " : "
            + (stat_prop_result_coverage["n_optional"] + stat_prop_result_coverage["n_required"]).astype(str)
            + " / "
            + calc_coverage_percentage()
        )
        #   + (stat_prop_result_coverage["n_optional"]+stat_prop_result_coverage["n_required"]).astype(str)+

        stat_prop_result_coverage.insert(
            loc=1,
            column=(
                " Summary Test Results Coverage: (No. Existing Required : No. Required / "
                "No. Optional Existing : No Optional / No. Total Existing Test : No. Total Tests) % "
            ),
            value=stat_prop_coverage_summary,
        )

    final_status_results = (
        stat_prop_result_all.drop("n_success", axis=1)
        .drop("n_failed", axis=1)
        .drop("n_skipped", axis=1)
        .drop("n_total", axis=1)
    )

    if stat_prop_result_cv_test is not None and len(stat_prop_result_cv_test) > 0:
        final_test_results = (
            stat_prop_result_cv_test.drop("n_pass", axis=1).drop("n_fail", axis=1).drop("n_tests", axis=1)
        )
        final_test_table = pd.merge(
            final_status_results, final_test_results, on=["model", "enc_format", "opcode"], how="inner"
        )
    else:
        final_test_table = final_status_results
    if stat_prop_result_coverage is not None and len(stat_prop_result_coverage) > 0:
        final_coverage_results = (
            stat_prop_result_coverage.drop("n_exists", axis=1)
            .drop("n_required", axis=1)
            .drop("n_optional", axis=1)
            .drop("n_required_exists", axis=1)
            .drop("n_optional_exists", axis=1)
        )
        final_coverage_table = pd.merge(
            final_test_table, final_coverage_results, on=["model", "enc_format", "opcode"], how="inner"
        )
    else:
        final_coverage_table = final_test_table

    save_data_frames_as_html_to_file(final_test_table, "Grouped_stat_prop_result_test.html")
    save_data_frames_as_html_to_file(final_coverage_table, "Grouped_stat_prop_result_cv.html")
    save_data_frames_as_html_to_file(stat_prop_result_cv_df, "Grouped_stat_prop_result_all.html")

    final_test_table = final_test_table.drop("instr", axis=1)
    final_coverage_table = final_coverage_table.drop("instr", axis=1)
    save_data_frames_as_html_to_file(final_test_table, "Grouped_stat_prop_result_test_wo_instr.html")
    save_data_frames_as_html_to_file(final_coverage_table, "Grouped_stat_prop_result_cv_wo_instr.html")


#   print("FINAL TEST_RESULTS\n")
#   print(final_test_results)

#    print('FINALE STATUS RESULTS\n')
#    print(final_status_results)


#    print("Final Test Table\n")
#    print(final_test_table)
#    print('finale Coverage Results\n')
#    print(final_coverage_results)
#    print("Final Coverage Table\n")
#    print(final_coverage_table)

if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel by removing unused constants."""

import sys
import argparse
import logging
import pathlib

import pandas as pd

from m2isar.metamodel import patch_model

from seal5.model import Seal5Constraint
from seal5.model_utils import load_model, dump_model

from . import visitor

from seal5.logging import Logger

logger = Logger("transform.collect_raises")


class VisitorContext:
    def __init__(self):
        self.cond_stack = []
        self.raises = []
        self.found_raise = False


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--compat", action="store_true")
    return parser


def run(args):
    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_instructions": [],
        "failed_instructions": [],
        "success_instructions": [],
    }
    for _, set_def in model_obj.sets.items():
        metrics["n_sets"] += 1
        logger.debug("collecting raises for set %s", set_def.name)
        patch_model(visitor)
        for _, instr_def in set_def.instructions.items():
            metrics["n_instructions"] += 1
            context = VisitorContext()
            logger.debug("collecting raises for instr %s", instr_def.name)
            try:
                context = VisitorContext()
                instr_def.operation.generate(context)
                if len(instr_def.constraints) > 0:
                    raise NotImplementedError("Can not yet append existing constraints")
                constraints = []
                for x in context.raises:
                    # print("x", x, type(x), dir(x))
                    mode_map = ["Exception", "Interrupt"]
                    mode = mode_map[x[1][0]]
                    assert mode != "Interrupt", "Interrupts not supported"
                    exec_map = [
                        "Instruction address misaligned",
                        "Instruction_access_fault",
                        "Illegal instruction",
                        "Breakpoint",
                        "Load address misaligned",
                        "Load access fault",
                        "Store/AMO address misaligned",
                        "Store/AMO access fault",
                        "Environment call from U-mode",
                        "Environment call from S-mode",
                        "Reserved",
                        "Environment call from M-mode",
                        "Instruction page fault",
                        "Load page fault",
                        "Reserved",
                        "Store/AMO page fault",
                    ]
                    code = x[1][1]
                    assert code < len(exec_map), "Out of bounds"
                    text = exec_map[code]
                    description = f"{mode}({code}): {text}"
                    # print("description", description)
                    # input("qqq2")
                    constraint = Seal5Constraint([x[0]], description=description)
                    constraints.append(constraint)

                instr_def.constraints = constraints
                # print("context.raises", context.raises)
                # input("next?")
                metrics["n_success"] += 1
                metrics["success_instructions"].append(instr_def.name)
            except Exception as ex:
                logger.exception(ex)
                metrics["n_failed"] += 1
                metrics["failed_instructions"].append(instr_def.name)

    dump_model(model_obj, out_path, compat=args.compat)
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Seal5 utility functions."""
import os
import sys
import shutil

# import distutils
import subprocess
import multiprocessing
from pathlib import Path
from typing import List, Callable, Optional


def get_logger():
    from seal5.logging import Logger

    return Logger("utils")


def str2bool(value, allow_none=False):
    if value is None:
        assert allow_none, "str2bool received None value while allow_none=False"
        return value
    if isinstance(value, (int, bool)):
        return bool(value)
    assert isinstance(value, str), "Expected str value"
    # return bool(distutils.util.strtobool(value))
    value = value.strip()
    if value.isdigit():
        return bool(int(value))
    value = value.lower()
    if value in ["true", "on", "yes"]:
        return True
    if value in ["false", "off", "no"]:
        return False
    if len(value) == 0:
        return False
    assert False, f"Unhandled case: {value}"


def copy(src, dest):
    shutil.copy(src, dest)


def is_populated(path):
    if not isinstance(path, Path):
        path = Path(path)
    return path.is_dir() and os.listdir(path.resolve())


def exec_getout(
    *args: List[str],
    ignore_output: bool = False,
    live: bool = False,
    print_func: Callable = print,
    handle_exit=None,
    err_func: Optional[Callable] = None,
    **kwargs,
) -> str:
    """Wrapper for running a program in a subprocess.

    Parameters
    ----------
    args : list
        The actual command.
    ignore_output : bool
        Do not get the stdout and stderr or the subprocess.
    live : bool
        Print the output line by line instead of only at the end.
    print_func : Callable
        Function which should be used to print sysout messages.
    err_func : Callable
        Function which should be used to print errors.
    kwargs: dict
        Arbitrary keyword arguments passed through to the subprocess.

    Returns
    -------
    out : str
        The command line output of the command
    """

    if err_func is None:
        logger = get_logger()  # instantiate here
        err_func = logger.error
        logger.debug("- Executing: %s", str(args))
    else:
        err_func = print

    if ignore_output:
        assert not live
        subprocess.run(args, **kwargs, check=True)
        return None

    out_str = ""
    if live:
        with subprocess.Popen(
            args,
            **kwargs,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        ) as process:
            for line in process.stdout:
                new_line = line.decode(errors="replace")
                out_str = out_str + new_line
                print_func(new_line.replace("\n", ""))
            exit_code = None
            while exit_code is None:
                exit_code = process.poll()
            if handle_exit is not None:
                exit_code = handle_exit(exit_code, out_str)
            if exit_code != 0:
                err_func(out_str)
            assert exit_code == 0, "The process returned an non-zero exit code {}! (CMD: `{}`)".format(
                exit_code, " ".join(list(map(str, args)))
            )
    else:
        with subprocess.Popen(args, **kwargs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as p:
            out_str = p.communicate()[0].decode(errors="replace")
            exit_code = p.poll()
        print_func(out_str)
        if handle_exit is not None:
            exit_code = handle_exit(exit_code, out_str)
        if exit_code != 0:
            err_func(out_str)
        assert exit_code == 0, "The process returned an non-zero exit code {}! (CMD: `{}`)".format(
            exit_code, " ".join(list(map(str, args)))
        )

    return out_str


def get_cmake_args(cfg: dict):
    ret = []
    for key, value in cfg.items():
        if isinstance(value, bool):
            value = "ON" if value else "OFF"
        elif isinstance(value, list):
            value = ";".join(value)
        else:
            assert isinstance(value, (int, str)), "Unsupported cmake cfg"
        ret.append(f"-D{key}={value}")
    return ret


def cmake(src, *args, debug: Optional[bool] = None, use_ninja: Optional[bool] = None, cwd=None, **kwargs):
    if cwd is None:
        raise RuntimeError("Please always pass a cwd to cmake()")
    if isinstance(cwd, Path):
        cwd = str(cwd.resolve())
    extra_args = []
    if not any("CMAKE_BUILD_TYPE" in x for x in args):
        if debug is not None:
            build_type = "Debug" if debug else "Release"
        extra_args.append("-DCMAKE_BUILD_TYPE=" + build_type)
    if use_ninja:
        extra_args.append("-GNinja")
    cmd = ["cmake", "-B", cwd, "-S", str(src)] + extra_args + list(args)
    return exec_getout(*cmd, cwd=cwd, **kwargs)


def make(*args, target=None, threads=multiprocessing.cpu_count(), use_ninja=False, cwd=None, verbose=False, **kwargs):
    del verbose  # unused
    if cwd is None:
        raise RuntimeError("Please always pass a cwd to make()")
    if isinstance(cwd, Path):
        cwd = str(cwd.resolve())
    # TODO: make sure that ninja is installed?
    extra_args = []
    if target:
        extra_args.append(target)
    tool = "ninja" if use_ninja else "make"
    extra_args.append("-j" + str(threads))
    cmd = [tool] + extra_args + list(args)
    return exec_getout(*cmd, cwd=cwd, **kwargs)


def python(*args, **kwargs):
    """Run a python script with the current interpreter."""
    return exec_getout(sys.executable, *args, **kwargs)


def clean_path(path: Path, interactive: bool = False):
    if interactive:
        answer = input(f"Remove '{path}' [Y|n]")
        if len(answer) == 0:
            answer = "Y"
        allow = answer.lower() in ["j", "y"]
    else:
        allow = True
    if not path.is_dir():
        # TODO: warning
        return
    if allow:
        shutil.rmtree(path)


def merge_dicts(a: dict, b: dict, path: Optional[list] = None):
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dicts(a[key], b[key], path + [str(key)])
            elif a[key] != b[key]:
                assert type(a[key]) is type(b[key])
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def ask_user(
    text, default: bool, yes_keys: Optional[list] = None, no_keys: Optional[list] = None, interactive: bool = True
):
    if yes_keys is None:
        yes_keys = ["y", "j"]
    if no_keys is None:
        no_keys = ["n"]

    assert len(yes_keys) > 0 and len(no_keys) > 0
    if not interactive:
        return default
    if default:
        suffix = " [{}/{}] ".format(yes_keys[0].upper(), no_keys[0].lower())
    else:
        suffix = " [{}/{}] ".format(yes_keys[0].lower(), no_keys[0].upper())
    message = text + suffix
    answer = input(message)
    if default:
        return answer.lower() not in no_keys and answer.upper() not in no_keys
    return not (answer.lower() not in yes_keys and answer.upper() not in yes_keys)


def is_power_of_two(n):
    return (n & (n - 1) == 0) and n != 0


def parse_cond(cond, settings=None):
    # TODO: is this secure enough?
    locals_ = {"settings": settings} if settings is not None else {}
    res = eval(cond, {"__builtins__": {}}, locals_)
    return res



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import argparse
import logging
import pathlib

from seal5.index import NamedPatch, write_index_yaml
from seal5.settings import RISCVLegalizerSettings, Seal5Settings

from seal5.logging import Logger

logger = Logger("backends.riscv_gisel_legalizer")

# if (ST.hasVendorXCvsimd()) {


def type_helper(ty):
    """Converts seal5-specific types into LLVMs LLT/VTs."""
    ty_ = ty.replace("seal5_", "")
    if ty_.startswith("p"):
        sz = ty_[1:]
        sz = int(sz)
        return f"const LLT {ty} = LLT::pointer({sz}, XLen);"
    if ty_.startswith("s"):
        sz = ty_[1:]
        sz = int(sz)
        return f"const LLT {ty} = LLT::scalar({sz});"
    if ty_.startswith("v"):
        n, sz = ty_[1:].split("i", 1)
        n = int(n)
        sz = int(sz)
        return f"const LLT {ty} = LLT::fixed_vector({n}, LLT::scalar({sz}));"
    raise RuntimeError(f"Unsupported: {ty}")


def gen_riscv_gisel_legalizer_str(legalizer_settings: RISCVLegalizerSettings):
    """Generate GISel legalizer code."""
    ops = legalizer_settings.ops
    used_types = []
    types_lines = []
    settings_lines = []
    if ops is None:
        return ""
    for op in ops:
        # print("op", op)
        names = op.name
        # print("names", names)
        types = op.types
        # print("types", types)
        onlyif = op.onlyif
        # print("onlyif", onlyif)
        if not isinstance(names, list):
            assert isinstance(names, str)
            names = [names]
        # print("names'", names)
        if not isinstance(types, list):
            assert isinstance(types, str)
            names = [names]
        # print("types'", types)
        types = [f"seal5_{ty}" for ty in types]
        types_str = "{" + ", ".join(types) + "}"
        for ty in types:
            if ty not in used_types:
                line = type_helper(ty)
                types_lines.append(line)
                used_types.append(ty)
        assert len(names) > 0
        if len(names) == 1:
            names_str = names[0]
        else:  # TODO: iterate over ops!
            # names_str = "{" + ", ".join(names) + "}"
            raise NotImplementedError
        line = ""
        if onlyif:
            cond = " && ".join(["ST." + pred[0].lower() + pred[1:] + "()" for pred in onlyif])
            line += f"if ({cond}) "
        line += f"getActionDefinitionsBuilder({names_str}).legalFor({types_str});"
        settings_lines.append(line)

    # print("used_types")
    ret = ""
    ret += "{\n"
    ret += "\n".join(types_lines)
    ret += "\n"
    ret += "\n".join(settings_lines)
    ret += "\n}"
    return ret


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--index", default=None, help="Output index to file")
    parser.add_argument("--ext", type=str, default="td", help="Default file extension (if using --splitted)")
    parser.add_argument("--yaml", type=str, default=None)
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    model_name = top_level.stem
    # abs_top_level = top_level.resolve()

    # is_seal5_model = False
    # # print("top_level", top_level)
    # # print("suffix", top_level.suffix)
    # if top_level.suffix == ".seal5model":
    #     is_seal5_model = True
    # if args.output is None:
    #     assert top_level.suffix in [".m2isarmodel", ".seal5model"], "Can not infer model type from file extension."
    #     raise NotImplementedError

    #     # out_path = top_level.parent / (top_level.stem + ".core_desc")
    # else:
    assert args.output is not None
    out_path = pathlib.Path(args.output)

    assert args.yaml is not None
    assert pathlib.Path(args.yaml).is_file()
    settings = Seal5Settings.from_yaml_file(args.yaml)

    # logger.info("loading models")
    # if not is_seal5_model:
    #     raise NotImplementedError

    # load models

    artifacts = {}
    artifacts[None] = []  # used for global artifacts
    if settings:
        riscv_settings = settings.riscv
        model_settings = settings.models.get(model_name)
        if model_settings is not None:
            model_riscv_settings = model_settings.riscv
            if model_riscv_settings is not None:
                riscv_settings = riscv_settings.merge(model_riscv_settings)

        if riscv_settings:
            legalization_settings = riscv_settings.legalization
            if legalization_settings:
                gisel_settings = legalization_settings.get("gisel", None)
                if gisel_settings:
                    content = gen_riscv_gisel_legalizer_str(gisel_settings)
                    if len(content) > 0:
                        with open(out_path, "w", encoding="utf-8") as f:
                            f.write(content)
                        riscv_gisel_legalizer_patch = NamedPatch(
                            "llvm/lib/Target/RISCV/GISel/RISCVLegalizerInfo.cpp",
                            key="riscv_legalizer_info",
                            src_path=out_path,
                        )
                        artifacts[None].append(riscv_gisel_legalizer_patch)
    if args.metrics:
        raise NotImplementedError
        # metrics_file = args.metrics
        # with open(metrics_file, "w") as f:
        #     f.write(",".join(metrics.keys()))
        #     f.write("\n")
        #     f.write(",".join(map(str, metrics.values())))
        #     f.write("\n")
    if args.index:
        if sum(map(len, artifacts.values())) > 0:
            global_artifacts = artifacts.get(None, [])
            set_artifacts = {key: value for key, value in artifacts.items() if key is not None}
            index_file = args.index
            write_index_yaml(index_file, global_artifacts, set_artifacts, content=True)
        else:
            logger.warning("No patches generated. No index file will be written.")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Remove (rd != 0) checks from M2-ISA-R/Seal5 metamodel."""

import sys
import argparse
import logging
import pathlib

from seal5.settings import Seal5Settings, ExtensionsSettings, RISCVSettings
from seal5.model_utils import load_model, dump_model

from seal5.logging import Logger

logger = Logger("transform.process_settings")


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--yaml", type=str, default=None)
    return parser


def run(args):
    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output
    model_name = top_level.stem

    model_obj = load_model(top_level, compat=False)

    # load settings
    if args.yaml is None:
        raise RuntimeError("Undefined --yaml not allowed")
    settings = Seal5Settings.from_yaml_file(args.yaml)

    if model_obj.settings is None:
        model_obj.settings = settings
    else:
        model_obj.settings.merge(settings, overwrite=True, inplace=True)

    for set_name, set_def in model_obj.sets.items():
        model_settings = settings.models.get(model_name)
        is_group_set = False
        if len(set_def.instructions) == 0:
            assert len(set_def.extension) > 0
            is_group_set = True
        ext_settings = None
        if model_settings is not None:
            ext_settings = model_settings.extensions.get(set_name, None)
        if ext_settings is None:
            ext_settings = ExtensionsSettings()
            ext_settings.feature = ext_settings.get_feature(set_name)
        if not is_group_set:
            riscv_settings = ext_settings.riscv
            if riscv_settings is None:
                riscv_settings = RISCVSettings(xlen=set_def.xlen)
            if riscv_settings.xlen is None:
                riscv_settings.xlen = set_def.xlen
        if set_def.settings is None:
            set_def.settings = ext_settings  # TODO: decide how to do this properly
        else:
            set_def.settings.merge(ext_settings, overwrite=True, inplace=True)

    dump_model(model_obj, out_path)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Properties report writer for Seal5."""

import argparse
import logging
import pathlib

import pandas as pd

from seal5.model import Seal5OperandAttribute, Seal5InstrAttribute
from seal5.model_utils import load_model
from seal5.riscv_utils import detect_opcode, detect_format

from seal5.logging import Logger

logger = Logger("backends.properties_writer")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", nargs="+", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--fmt", type=str, choices=["auto", "csv", "pkl", "md"], default="auto")
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    assert args.output is not None
    out_path = pathlib.Path(args.output)

    # resolve model paths
    properties_data = []
    top_levels = args.top_level
    for top_level in top_levels:
        top_level = pathlib.Path(top_level)

        model_obj = load_model(top_level, compat=args.compat)

        # preprocess model
        # print("model", model)
        for set_name, set_def in model_obj.sets.items():
            # print("set_name", set_name)
            if len(set_def.instructions) == 0:
                continue
            xlen = set_def.xlen
            model = top_level.stem

            def get_set_properties(set_name, xlen):
                is_rv32 = xlen == 32
                is_rv64 = xlen == 64
                return {
                    "set": set_name,
                    "xlen": xlen,
                    "is_rv32": is_rv32,
                    "is_rv64": is_rv64,
                }

            for instr_def in set_def.instructions.values():
                instr_name = instr_def.name
                # print("instr_name", instr_name)
                operands = instr_def.operands

                def get_enc_properties(instr_def):
                    enc_size = instr_def.size
                    is_compressed = enc_size == 16
                    # TODO: opcode
                    # TODO: format

                    # opcode, _ = detect_opcode(instr_def)
                    # enc_format, _ = detect_format(instr_def)
                    opcode = instr_def.attributes.get(Seal5InstrAttribute.OPCODE_NAME)
                    if opcode is None:
                        opcode, _ = detect_opcode(instr_def)
                    else:
                        opcode = opcode.value
                    enc_format = instr_def.attributes.get(Seal5InstrAttribute.ENC_FORMAT)
                    if enc_format is None:
                        enc_format, _ = detect_format(instr_def)
                    else:
                        enc_format = enc_format.value

                    return {
                        "enc_size": enc_size,
                        "is_compressed": is_compressed,
                        "enc_format": enc_format,
                        "opcode": opcode,
                    }

                def get_operands_properties(operands):
                    num_operands = len(operands)
                    num_inputs = len([op for op in operands.values() if Seal5OperandAttribute.IN in op.attributes])
                    num_outputs = len([op for op in operands.values() if Seal5OperandAttribute.OUT in op.attributes])
                    num_inouts = len([op for op in operands.values() if Seal5OperandAttribute.INOUT in op.attributes])
                    num_regs = len([op for op in operands.values() if Seal5OperandAttribute.IS_REG in op.attributes])
                    num_gprs = len(
                        [
                            op
                            for op in operands.values()
                            if Seal5OperandAttribute.IS_REG in op.attributes
                            and op.attributes[Seal5OperandAttribute.REG_CLASS] == "GPR"
                        ]
                    )
                    num_imms = len([op for op in operands.values() if Seal5OperandAttribute.IS_IMM in op.attributes])
                    imm_types = set(
                        op.attributes[Seal5OperandAttribute.TYPE]
                        for op in operands.values()
                        if Seal5OperandAttribute.IS_IMM in op.attributes
                    )
                    has_imm_leaf = any(
                        Seal5OperandAttribute.IS_IMM_LEAF in op.attributes
                        for op in operands.values()
                        if Seal5OperandAttribute.IS_IMM in op.attributes
                    )
                    is_multi_in = num_inputs + num_inouts > 1
                    is_single_in = num_inputs + num_inouts == 1
                    is_multi_out = num_outputs + num_inouts > 1
                    is_single_out = num_outputs + num_inouts == 1
                    is_mimo = is_multi_in and is_multi_out
                    is_miso = is_multi_in and is_single_out
                    is_siso = is_single_in and is_single_out
                    is_simo = is_single_in and is_multi_out
                    return {
                        "num_operands": num_operands,
                        "num_inputs": num_inputs,
                        "num_outputs": num_outputs,
                        "num_inouts": num_inouts,
                        "num_regs": num_regs,
                        "num_gprs": num_gprs,
                        "num_imms": num_imms,
                        "imm_types": imm_types,
                        "is_multi_in": is_multi_in,
                        "is_single_in": is_single_in,
                        "is_multi_out": is_multi_out,
                        "is_single_out": is_single_out,
                        "is_mimo": is_mimo,
                        "is_miso": is_miso,
                        "is_siso": is_siso,
                        "is_simo": is_simo,
                        "has_imm_leaf": has_imm_leaf,
                        # TODO: num_csrs, num_fprs, num_custom...
                    }

                def get_side_effect_properties(attrs):
                    uses_custom_reg = Seal5InstrAttribute.USES in attrs and len(attrs[Seal5InstrAttribute.USES]) > 0
                    defs_custom_reg = Seal5InstrAttribute.DEFS in attrs and len(attrs[Seal5InstrAttribute.DEFS]) > 0
                    return {
                        "has_side_effects": Seal5InstrAttribute.HAS_SIDE_EFFECTS in attrs,
                        "may_load": Seal5InstrAttribute.MAY_LOAD in attrs,
                        "may_store": Seal5InstrAttribute.MAY_STORE in attrs,
                        "is_terminator": Seal5InstrAttribute.IS_TERMINATOR in attrs,
                        "is_branch": Seal5InstrAttribute.IS_BRANCH in attrs,
                        "skip_pattern_gen": Seal5InstrAttribute.SKIP_PATTERN_GEN in attrs,
                        "uses_custom_reg": uses_custom_reg,
                        "defs_custom_reg": defs_custom_reg,
                    }

                attrs = instr_def.attributes
                data = {
                    "model": model,
                    **get_set_properties(set_name, xlen),
                    "instr": instr_name,
                    **get_enc_properties(instr_def),
                    **get_operands_properties(operands),
                    **get_side_effect_properties(attrs),
                    # TODO: operations
                }
                properties_data.append(data)
    properties_df = pd.DataFrame(properties_data)
    fmt = args.fmt
    if fmt == "auto":
        fmt = out_path.suffix
        assert len(fmt) > 1
        fmt = fmt[1:].lower()

    if fmt == "csv":
        properties_df.to_csv(out_path, index=False)
    elif fmt == "pkl":
        properties_df.to_pickle(out_path)
    elif fmt == "md":
        # properties_df.to_markdown(out_path, tablefmt="grid", index=False)
        properties_df.to_markdown(out_path, index=False)
    else:
        raise ValueError(f"Unsupported fmt: {fmt}")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2024
# Chair of Electrical Design Automation
# Technical University of Munich

"""This module inlines functions within instructions in the M2-ISA-R metamodel."""

from m2isar.metamodel import arch, behav


from seal5.model import Seal5FunctionAttribute

# pylint: disable=unused-argument


class ReplaceContext:
    def __init__(self, mapping):
        self.mapping = mapping


def operation(self: behav.Operation, context):
    statements = []
    for stmt in self.statements:
        try:
            temp = stmt.generate(context)
            if isinstance(temp, list):
                statements.extend(temp)
            else:
                statements.append(temp)
        except (NotImplementedError, ValueError):
            print(f"cant simplify {stmt}")

    self.statements = statements
    return self


def binary_operation(self: behav.BinaryOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def slice_operation(self: behav.SliceOperation, context):
    self.expr = self.expr.generate(context)
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def concat_operation(self: behav.ConcatOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def number_literal(self: behav.IntLiteral, context):
    return self


def int_literal(self: behav.IntLiteral, context):
    return self


def scalar_definition(self: behav.ScalarDefinition, context):
    return self


def break_(self: behav.Break, context):
    return self


def assignment(self: behav.Assignment, context):
    self.target = self.target.generate(context)
    self.expr = self.expr.generate(context)

    return self


def conditional(self: behav.Conditional, context):
    self.conds = [x.generate(context) for x in self.conds]
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def loop(self: behav.Loop, context):
    self.cond = self.cond.generate(context)
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def ternary(self: behav.Ternary, context):
    self.cond = self.cond.generate(context)
    self.then_expr = self.then_expr.generate(context)
    self.else_expr = self.else_expr.generate(context)

    return self


def return_(self: behav.Return, context):
    if self.expr is not None:
        self.expr = self.expr.generate(context)

    return self


def unary_operation(self: behav.UnaryOperation, context):
    self.right = self.right.generate(context)

    return self


def named_reference(self: behav.NamedReference, context):
    # print("named_refernce", self, dir(self))
    if isinstance(context, ReplaceContext):
        matched = context.mapping.get(self.reference.name, None)
        if matched is not None:
            return matched
    return self


def indexed_reference(self: behav.IndexedReference, context):
    self.index = self.index.generate(context)

    return self


def type_conv(self: behav.TypeConv, context):
    # print("type_conv", self)
    self.expr = self.expr.generate(context)

    return self


def callable_(self: behav.Callable, context):
    # print("callable_", self, dir(self))
    # print("self.ref_or_name", self.ref_or_name)
    # print("context.functions", context.functions)
    if isinstance(self.ref_or_name, str):
        function_def = context.functions.get(self.ref_or_name)
    else:
        assert isinstance(self.ref_or_name, arch.Function)
        function_def = self.ref_or_name
    # Retrieve the function definition
    # print("function_def", function_def)
    # print("function_def.attributes", function_def.attributes)
    if Seal5FunctionAttribute.INLINE not in function_def.attributes:
        return self

    # Check if the function is external or has early returns that prevent inlining
    if function_def is None or function_def.extern:
        return self  # Cannot inline an external or undefined function

    if len(function_def.operation.statements) != 1:
        return self

    statement = function_def.operation.statements[0]
    ret_dtype = function_def.data_type
    ret_size = None  # TODO: add sized return types to m2isar functions
    statement = behav.TypeConv(ret_dtype, ret_size, statement)
    # print("statement", statement)
    while isinstance(statement, behav.Block):
        # print("BLOCK", statement, dir(statement))
        statements = statement.statements
        if len(statements) != 1:
            return self
        statement = statements[0]
    # print("statement", statement, dir(statement))
    # print("?", [isinstance(stmt, behav.Return) for stmt in function_def.operation.statements])
    if not isinstance(statement, behav.Return):
        return self
    # return self  # Skip inlining if the function has a return statement
    # Map the arguments of the function to the corresponding expressions
    arg_map = {param_name: arg for param_name, arg in zip(function_def.args, self.args)}
    # print("arg_map", arg_map)
    ret_expr = statement.expr
    # print("ret_expr", ret_expr)
    replace_context = ReplaceContext(mapping=arg_map)
    ret_expr.generate(replace_context)
    ret_expr.generate(context)
    return ret_expr

    return self


def group(self: behav.Group, context):
    self.expr = self.expr.generate(context)

    return self



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""TODO"""

from m2isar.metamodel import arch, behav
from seal5 import model

# pylint: disable=unused-argument


def operation(self: behav.Operation, context):
    statements = []
    for stmt in self.statements:
        temp = stmt.generate(context)
        if isinstance(temp, list):
            statements.extend(temp)
        else:
            statements.append(temp)

    self.statements = statements
    return self


def binary_operation(self: behav.BinaryOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def slice_operation(self: behav.SliceOperation, context):
    self.expr = self.expr.generate(context)
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def concat_operation(self: behav.ConcatOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def number_literal(self: behav.IntLiteral, context):
    return self


def int_literal(self: behav.IntLiteral, context):
    return self


def scalar_definition(self: behav.ScalarDefinition, context):
    return self


def break_(self: behav.Break, context):
    return self


def assignment(self: behav.Assignment, context):
    self.target = self.target.generate(context)
    self.expr = self.expr.generate(context)

    return self


def conditional(self: behav.Conditional, context):
    # print("conditional")
    self.conds = [x.generate(context) for x in self.conds]
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def loop(self: behav.Loop, context):
    self.cond = self.cond.generate(context)
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def ternary(self: behav.Ternary, context):
    self.cond = self.cond.generate(context)
    self.then_expr = self.then_expr.generate(context)
    self.else_expr = self.else_expr.generate(context)

    return self


def return_(self: behav.Return, context):
    if self.expr is not None:
        self.expr = self.expr.generate(context)

    return self


def unary_operation(self: behav.UnaryOperation, context):
    self.right = self.right.generate(context)

    return self


def named_reference(self: behav.NamedReference, context):
    # if isinstance(self.reference, arch.BitFieldDescr):
    # raise NotImplementedError
    return self


def indexed_reference(self: behav.IndexedReference, context):
    self.index = self.index.generate(context)

    return self


def type_conv(self: behav.TypeConv, context):
    # print("type_conv", self, dir(self))
    self.expr = self.expr.generate(context)
    if isinstance(self.expr, behav.NamedReference):  # imm?
        if isinstance(self.expr.reference, arch.BitFieldDescr):
            name = self.expr.reference.name
            assert name in context.operands
            op = context.operands[name]
            assert isinstance(op, model.Seal5ImmOperand)
            ty = op.ty
            width = ty.width
            if ty != self.data_type:
                # update
                if ty.datatype == arch.DataType.U:
                    ty.datatype = self.data_type
                    op.ty = ty

                else:
                    assert False, "Conflicting types"
            if self.size is not None:
                if width != self.size:
                    if self.size < width:  # trunc!
                        assert False, "truncation not allowed here"
                    elif self.size > width:  # zext/sext!
                        assert False, "sign/zero extension not allowed here"
            # print("op", name, op, type(op))
            # if not isinstance(op, model.Seal5ImmOperand):
            # op = model.Seal5ImmOperand(op.name, op.ty, op.attributes, op.constraints)
            context.operands[name] = op
            # if self.data_type != arch.DataType.U:  # Default to unsigned
            # if self.size is not None:
            #     # Do not allow truncation
            #     assert self.size == self.expr.reference.size
            # else:
            #     # add explicit size (optional?)
            #     self.size = self.expr.reference.size
            # if self.data_type:
            #     if self.data_type != self.reference.data_type:
            #         # Do not update BitFieldDescr for now...
            #         # self.expr.reference.data_type = self.data_type
    elif isinstance(self.expr, behav.IndexedReference):  # X[reg]?
        if isinstance(self.expr.reference, arch.Memory):
            if isinstance(self.expr.index, behav.NamedReference):
                if isinstance(self.expr.index.reference, arch.BitFieldDescr):
                    op_name = self.expr.index.reference.name
                    assert op_name in context.operands
                    op = context.operands[op_name]
                    assert isinstance(op, model.Seal5RegOperand)
                    reg_ty = op.reg_ty
                    width = reg_ty.width
                    # print("op", op, op.name, op.reg_class, op.reg_ty, op.ty, op.attributes, op.constraints)
                    if reg_ty != self.data_type:
                        # update
                        if reg_ty.datatype == arch.DataType.U:
                            reg_ty.datatype = self.data_type
                            op.reg_ty = reg_ty

                        else:
                            assert False, "Conflicting types"
                    if self.size is not None:
                        if width != self.size:
                            if self.size < width:  # trunc!
                                return self
                                # assert False, "truncation not allowed here"
                            if self.size > width:  # zext/sext!
                                assert False, "sign/zero extension not allowed here"
                    context.operands[op_name] = op

    return self


def callable_(self: behav.Callable, context):
    self.args = [stmt.generate(context) for stmt in self.args]

    return self


def group(self: behav.Group, context):
    self.expr = self.expr.generate(context)

    return self


def procedure_call(self: behav.ProcedureCall, context):
    self.fn_args = [arg.generate(context) for arg in self.args]
    return self



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""This module contains the python files generated by ANTLR. These are version controlled
to allow users without a working ANTLR and/or Java setup to also use M2-ISA-R.
"""

from .CoreDSL2Lexer import CoreDSL2Lexer
from .CoreDSL2Listener import CoreDSL2Listener
from .CoreDSL2Parser import CoreDSL2Parser
from .CoreDSL2Visitor import CoreDSL2Visitor


# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import argparse
import logging
import pathlib
from typing import Optional

import pandas as pd

from mako.template import Template

from m2isar.metamodel import arch

from seal5.index import NamedPatch, File, write_index_yaml
from seal5.utils import is_power_of_two
from seal5.settings import IntrinsicDefn, ExtensionsSettings
from seal5.model_utils import load_model

# from seal5.settings import ExtensionsSettings

from .templates import template_dir

from seal5.logging import Logger

logger = Logger("backends.riscv_instr_info")


class Operand:
    def __init__(self, name, lower, upper):
        self.name = name
        # self._name = name
        self.lower = lower
        self.upper = upper

    @property
    def length(self):
        return (self.upper - self.lower + 1) if self.upper >= self.lower else (self.lower - self.upper + 1)

    # @property
    # def name(self):
    #   ty, name_ = self._name.split(":")
    #   if self.lower == 1:
    #       ty += "_lsb0"
    #   ret = f"{ty}:{name_}"
    #   return ret

    def __repr__(self):
        return f"Operand({self.name}, {self.lower}, {self.upper})"


class EncodingField:
    def __init__(self, name, start, length, value=None, extra=None):
        self.name = name
        self.start = start
        self.length = length

        self.value = value
        self.const = value is not None
        self.extra = extra if extra is not None else value

    def __repr__(self):
        return f"EncodingField({self.name}, {self.start}, {self.length}, {self.value}, {self.const})"


def process_encoding(enc):
    # print("get_encoding", enc)
    operands = {}
    for e in reversed(enc):
        if isinstance(e, arch.BitField):
            name = e.name
            if name in operands:
                op = operands[name]
                op.lower = min(e.range.lower, op.lower)
                op.upper = max(e.range.upper, op.upper)
            else:
                op = Operand(name, e.range.lower, e.range.upper)
                operands[name] = op
    fields = []
    start = 0
    for e in reversed(enc):
        if isinstance(e, arch.BitField):
            name = e.name
            assert name in operands
            op = operands[name]
            if op.length > e.range.length:
                if e.range.length == 1:
                    name = name + "{" + str(e.range.lower - op.lower) + "}"
                else:
                    name = name + "{" + str(e.range.upper - op.lower) + "-" + str(e.range.lower - op.lower) + "}"
            new = EncodingField(name, start, e.range.length)
            start += e.range.length
        elif isinstance(e, arch.BitVal):
            # lower = 0
            new = EncodingField(None, start, e.length, e.value)
            start += e.length
        else:
            assert False
        fields.insert(0, new)
    # print("fields", fields)
    # input(">")
    return operands.values(), fields


def write_riscv_instruction_info(
    name,
    real_name,
    asm_str,
    ins_str,
    outs_str,
    enc,
    fields,
    # operands,
    size,
    details_str,
    attrs: Optional[dict] = None,
    constraints: Optional[list] = None,
    formats=False,
    compressed_pat=None,
):
    if attrs is None:
        attrs = {}
    if constraints is None:
        constraints = []
    if not formats:
        raise NotImplementedError("!InstrFormat")
        # instr_template = Template(filename=str(template_dir / "instr_tablegen2.mako"))
    else:
        instr_template = Template(filename=str(template_dir / "instr_tablegen.mako"))

    operands, fields = process_encoding(enc)

    def attr_helper(value):
        return int(value) if isinstance(value, bool) else value

    attrs = {key: attr_helper(value) for key, value in attrs.items()}
    constraints_str = ", ".join(constraints)

    if len(ins_str) == 0 and len(outs_str) == 0:
        assert len(operands) == 0, "Could not resolve in/out operands"

    out_str = instr_template.render(
        name=name,
        real_name=real_name,
        size=size,
        asm_str=asm_str,
        ins_str=ins_str,
        outs_str=outs_str,
        sched_str=str([]),
        operands=operands,
        fields=fields,
        attrs=attrs,
        constraints_str=constraints_str,
    )
    if compressed_pat:
        out_str += f"\n{compressed_pat}"

    if len(details_str) > 0:
        out_str = (
            details_str
            + " in {\n"
            + "\n".join([("  " + line) if len(line) > 0 else line for line in out_str.split("\n")])
            + "\n} // "
            + details_str
            + "\n"
        )

    logger.info("writing TableGen for instruction %s", name)

    return out_str


def gen_riscv_instr_info_str(instr, set_def):
    # print("instr", instr)
    name = instr.name
    # operands = instr.operands
    size = instr.size
    # print("operands", operands)
    # reads = instr.llvm_reads
    # writes = instr.llvm_writes
    constraints = instr.llvm_constraints
    # print("reads", reads)
    # print("writes", writes)
    constraints = instr.llvm_constraints
    # print("constraints", constraints)
    # constraints_str = ", ".join(constraints)
    # attributes = instr.attributes
    # print("attributes", attributes)
    real_name = instr.mnemonic
    asm_str = instr.llvm_asm_str
    # print("asm_str", asm_str)
    ins_str = instr.llvm_ins_str
    # print("ins_str", ins_str)
    outs_str = instr.llvm_outs_str
    # print("outs_str", outs_str)
    details_str = ""
    fields = instr.fields
    # print("fields")
    encoding = instr.encoding
    # print("encoding")
    attrs = instr.llvm_attributes
    # constraints = instr.constraints
    # if len(constraints) > 0:
    #     raise NotImplementedError
    formats = True
    compressed_pat = instr.llvm_get_compressed_pat(set_def)
    tablegen_str = write_riscv_instruction_info(
        name,
        real_name,
        asm_str,
        ins_str,
        outs_str,
        encoding,
        fields,
        # operands,
        size,
        details_str,
        attrs=attrs,
        constraints=constraints,
        formats=formats,
        compressed_pat=compressed_pat,
    )
    return tablegen_str


def gen_intrinsic_pattern(instr, set_name, intrinsic: IntrinsicDefn, ext_settings: Optional[ExtensionsSettings] = None):
    assert ext_settings is not None
    arch_ = ext_settings.get_arch(name=set_name)
    pat = f"""class Pat_{instr.name}<SDPatternOperator OpNode, Instruction Inst>
: Pat<(OpNode {instr.llvm_ins_str}), (Inst {instr.llvm_ins_str})>;
def : Pat_{instr.name}<int_riscv_{arch_}_{intrinsic.intrinsic_name}, {instr.name}>;"""
    return pat


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--splitted", action="store_true", help="Split per set")
    parser.add_argument("--formats", action="store_true", help="Also generate instruction formats")
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--index", default=None, help="Output index to file")
    parser.add_argument("--ext", type=str, default="td", help="Default file extension (if using --splitted)")
    parser.add_argument(
        "--no-add-intrinsics",
        dest="add_intrinsics",
        default=True,
        action="store_false",
        help="Suppress patterns for intrinsic functions",
    )
    parser.add_argument("--ignore-failing", action="store_true", help="Do not crash in case of errors.")
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    out_path = pathlib.Path(args.output)

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_instructions": [],
        "failed_instructions": [],
        "success_instructions": [],
        "skipped_sets": [],
        "failed_sets": [],
        "success_sets": [],
    }
    # preprocess model
    # print("model", model)
    settings = model_obj.settings
    artifacts = {}
    artifacts[None] = []  # used for global artifacts
    if args.splitted:
        content = ""
        # errs = []
        for set_name, set_def in model_obj.sets.items():
            if len(set_def.instructions) == 0:
                continue
            metrics["n_sets"] += 1
            set_name_lower = set_name.lower()
            artifacts[set_name] = []
            xlen = set_def.xlen
            assert xlen is not None
            assert is_power_of_two(xlen)
            assert xlen % 8 == 0
            includes = []
            set_dir = out_path / set_name
            set_dir.mkdir(exist_ok=True)
            ext_settings = set_def.settings
            pred = None
            if ext_settings is not None:
                pred = "Has" + ext_settings.get_predicate(name=set_name)
            # TODO: check for GPRC and require HasStdExtCOrZca?
            # TODO: check for GPR32Pair and require HasGPR32Pair
            # TODO: check for GPR32V2/GPR32V4 and require HasGPR32V
            for _, instr_def in set_def.instructions.items():
                metrics["n_instructions"] += 1
                out_name = f"{instr_def.name}InstrInfo.{args.ext}"
                output_file = set_dir / out_name
                try:
                    metrics["n_success"] += 1
                    metrics["success_instructions"].append(instr_def.name)
                    content = gen_riscv_instr_info_str(instr_def, set_def)
                    if len(content) > 0:
                        if args.add_intrinsics and settings.intrinsics.intrinsics:
                            # TODO: intrinsics should be dict keyed by instr name
                            for intrinsic in settings.intrinsics.intrinsics:
                                if intrinsic.instr_name.casefold() in [
                                    instr_def.mnemonic.casefold(),
                                    instr_def.name.casefold(),
                                ]:
                                    content += gen_intrinsic_pattern(
                                        instr_def, set_name, intrinsic, ext_settings=ext_settings
                                    )
                        assert pred is not None
                        predicate_str = f"Predicates = [{pred}, IsRV{xlen}]"
                        content = f"let {predicate_str} in {{\n{content}\n}}"
                        with open(output_file, "w", encoding="utf-8") as f:
                            f.write(content)
                        instr_info_patch = File(
                            f"llvm/lib/Target/RISCV/seal5/{set_name}/{output_file.name}",
                            src_path=output_file,
                        )
                        artifacts[set_name].append(instr_info_patch)
                        inc = f"seal5/{set_name}/{output_file.name}"
                        includes.append(inc)
                except Exception as ex:
                    logger.exception(ex)
                    metrics["n_failed"] += 1
                    metrics["failed_instructions"].append(instr_def.name)
            if len(includes) > 0:
                includes_str = "\n".join([f'include "{inc}"' for inc in includes])
                set_td_includes_patch = NamedPatch(
                    f"llvm/lib/Target/RISCV/seal5/{set_name}.td",
                    key=f"{set_name_lower}_set_td_includes",
                    content=includes_str,
                )
                artifacts[set_name].append(set_td_includes_patch)
    else:
        raise NotImplementedError
    if not args.ignore_failing:
        n_failed = metrics["n_failed"]
        if n_failed > 0:
            failed = metrics["failed_instructions"]
            failing_str = ", ".join(failed)
            logger.error("%s intructions failed: %s", n_failed, failing_str)
            raise RuntimeError("Abort due to errors")
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)
    if args.index:
        if sum(map(len, artifacts.values())) > 0:
            global_artifacts = artifacts.get(None, [])
            set_artifacts = {key: value for key, value in artifacts.items() if key is not None}
            index_file = args.index
            write_index_yaml(index_file, global_artifacts, set_artifacts, content=True)
        else:
            logger.warning("No patches generated. No index file will be written.")


if __name__ == "__main__":
    main()



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""LLVM utils for seal5."""
import re
import os
import shutil
from pathlib import Path
from typing import List, Optional, Union

import git
from git import RemoteProgress
from tqdm import tqdm

from seal5 import utils
from seal5.logging import Logger
from seal5.tools.git import get_author_from_settings
from seal5.settings import GitSettings, CcacheSettings

logger = Logger("tools")


def lookup_ccache():
    candidates = ["sccache", "ccache"]
    for candidate in candidates:
        found = shutil.which(candidate)
        if found:
            return found
    return None  # Not found


def check_llvm_repo(path: Path):
    repo = git.Repo(path)
    if repo.is_dirty(untracked_files=True):
        # Filter out .seal5/ directory
        out = repo.git.status("--porcelain")
        lines = out.split("\n")
        dirty = []
        for line in lines:
            _, file = line.strip().split(" ", 1)
            if file == ".seal5/":
                continue
            dirty.append(file)
        if len(dirty) > 0:
            logger.debug("Dirty files in LLVM repository: %s", ", ".join(dirty))
            return False

    return True


class CloneProgress(RemoteProgress):
    def __init__(self):
        super().__init__()
        self.pbar = tqdm()

    def update(self, op_code, cur_count, max_count=None, message=""):
        self.pbar.total = max_count
        self.pbar.n = cur_count
        self.pbar.refresh()


def clone_llvm_repo(
    dest: Path,
    clone_url: str,
    ref: Optional[str] = None,
    refresh: bool = False,
    label: str = "default",
    git_settings: GitSettings = None,
    depth: Optional[int] = None,
    default_major_version: int = 18,
    progress: bool = True,  # TODO: default to False and expose to flow.py
):
    sha = None
    version_info = {}
    repo = None
    if dest.is_dir() and utils.is_populated(dest):
        if refresh:
            logger.debug("Refreshing LLVM repository: %s", dest)
            repo = git.Repo(dest)
            repo.remotes.origin.set_url(clone_url)
            repo.remotes.origin.fetch()
            if ref:
                repo.git.checkout(ref)
                if ref not in repo.tags:
                    repo.git.pull("origin", ref)
    else:
        logger.debug("Cloning LLVM repository: %s", clone_url)
        if progress:
            clone_progress = CloneProgress()
        else:
            clone_progress = None
        no_checkout = ref is not None
        branch = None
        depth = depth if depth is not None and depth >= 0 else None
        if depth is not None and ref is not None:
            # assert "llvmorg" in ref  # Needs to be a valid branch name, not a tag
            branch = ref
        repo = git.Repo.clone_from(
            clone_url, dest, no_checkout=no_checkout, branch=branch, depth=depth, progress=clone_progress
        )
        if ref:
            logger.debug("Checking out branch: %s", ref)
            repo.git.checkout(ref)
    # Set author in repo config
    author = get_author_from_settings(git_settings)
    with repo.config_writer() as writer:
        writer.set_value("user", "name", author.name)
        writer.set_value("user", "email", author.email)
    # repo.create_tag(f"seal5-{label}-base", "-f", author=author)
    repo.create_tag(f"seal5-{label}-base", "-f")
    # git describe --tags --match "llvmorg-[0-9]*.[0-9]*.[0-9]*"
    try:
        describe = repo.git.describe("--tags", "--match", "llvmorg-[0-9]*.[0-9]*.[0-9]*")
    except git.GitCommandError:
        describe = None
    if describe:
        splitted = describe.split("-", 3)
        base = splitted[0]
        assert base == "llvmorg"
        version = splitted[1]
        major, minor, patch = version.split(".")
        version_info["major"] = int(major)
        version_info["minor"] = int(minor)
        version_info["patch"] = int(patch)
        rest = splitted[2] if len(splitted) > 2 else ""
        if "rc" in rest:
            rc = rest.split("-", 1)[0][2:]
            version_info["rc"] = int(rc)
    else:
        if default_major_version is not None:
            version_info["major"] = default_major_version

    sha = repo.head.commit.hexsha
    return repo, sha, version_info


def build_llvm(
    src: Path,
    dest: Path,
    target: str = "all",
    debug: Optional[bool] = None,
    use_ninja: Optional[bool] = None,
    verbose: bool = False,
    cmake_options: Optional[dict] = None,
    install: bool = False,
    install_dir: Optional[Union[str, Path]] = None,
    ccache_settings: Optional[CcacheSettings] = None,
):
    if cmake_options is None:
        cmake_options = {}
    if install:
        assert install_dir is not None
        assert Path(install_dir).parent.is_dir()
        cmake_options["CMAKE_INSTALL_PREFIX"] = str(install_dir)
    env = os.environ.copy()
    if ccache_settings:
        if ccache_settings.enable:
            ccache_executable = ccache_settings.executable
            ccache_directory = ccache_settings.directory
            if ccache_executable is None:
                ccache_executable = "auto"

            if ccache_executable == "auto":
                ccache_executable = lookup_ccache()
                assert ccache_executable is not None, "Could not resolve ccache executable"
            cmake_options["CMAKE_C_COMPILER_LAUNCHER"] = ccache_executable
            cmake_options["CMAKE_CXX_COMPILER_LAUNCHER"] = ccache_executable
            if ccache_directory is not None:
                env["CCACHE_DIR"] = ccache_directory
    cmake_args = utils.get_cmake_args(cmake_options)
    dest.mkdir(exist_ok=True)
    utils.cmake(
        src / "llvm",
        *cmake_args,
        use_ninja=use_ninja,
        debug=debug,
        cwd=dest,
        env=env,
        print_func=logger.info if verbose else logger.debug,
        live=True,
    )
    if install:
        assert target is None
        target = "install"
    utils.make(
        target=target, cwd=dest, print_func=logger.info if verbose else logger.debug, live=True, use_ninja=use_ninja
    )


def test_llvm(base: Path, build_dir: Path, test_paths: Optional[List[str]] = None, verbose: bool = False):
    if test_paths is None:
        test_paths = []
    env = os.environ.copy()
    old_path = env["PATH"]
    env["PATH"] = f"{build_dir}/bin:{old_path}"
    lit_exe = build_dir / "bin" / "llvm-lit"
    passed_tests = []
    failing_tests = []
    for test_path in test_paths:

        def handler(_code, _out):
            return 0

        test_path = Path(test_path)
        if test_path.is_absolute():
            test_file = test_path
        else:
            test_file = base / test_path
            if not test_file.is_file():
                try:
                    test_path = test_path.relative_to(base)
                except ValueError as exc:
                    raise RuntimeError(
                        f"Relative test path needs to be relative to base ({base}): {test_path}"
                    ) from exc
                test_file = base / test_path
                assert test_file.is_file()
        out = utils.exec_getout(
            lit_exe,
            test_file,
            print_func=logger.info if verbose else logger.debug,
            live=True,
            env=env,
            handle_exit=handler,
        )
        failing = re.compile(r"FAIL: LLVM :: (.*) \(").findall(out)
        if len(failing) > 0:
            failing_tests.extend(failing)
        passed = re.compile(r"PASS: LLVM :: (.*) \(").findall(out)
        if len(passed) > 0:
            passed_tests.extend(passed)

    return passed_tests, failing_tests


def detect_llvm_imm_types(llvm_dir: Union[str, Path]):
    # TODO: detect which ones are ImmLeafs!
    def get_grep_cmd(prefix):
        return f"grep -r \"def {prefix}imm\" llvm/lib/Target/RISCV | cut -d':' -f2 | tr -s ' ' | sed -e \"s/def //g\""

    def _parse_output(output):
        output = output.strip()
        return set(map(lambda x: x.strip(), output.splitlines()))

    uimm_types = _parse_output(
        utils.exec_getout(get_grep_cmd("u"), shell=True, print_func=lambda *_: None, live=False, cwd=llvm_dir)
    )
    simm_types = _parse_output(
        utils.exec_getout(get_grep_cmd("s"), shell=True, print_func=lambda *_: None, live=False, cwd=llvm_dir)
    )

    imm_types = uimm_types | simm_types
    return imm_types



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Seal5 dependencies."""

from typing import Optional, List, Union
from pathlib import Path

import git

from seal5.logging import Logger
from seal5.settings import LLVMVersion
from seal5.utils import is_populated
from .tools.llvm import CloneProgress  # TODO: move to other file

logger = Logger("dependencies")


class Dependency:
    pass


class GitDependency(Dependency):
    def __init__(self, name: str, clone_url: str, ref: Optional[str] = None, recursive: bool = False):
        self.name: str = name
        self.clone_url: str = clone_url
        self.ref: Optional[str] = ref
        self.recursive: bool = recursive

    def clone(
        self,
        dest: Path,
        overwrite: bool = False,
        depth: Optional[int] = None,
        progress: bool = False,
        sparse: bool = False,
        sparse_filter=None,
    ):
        del overwrite  # TODO: rename to override
        if is_populated(dest):
            logger.debug("Updating repository: %s", dest)
            repo = git.Repo(dest)
            repo.remotes.origin.set_url(self.clone_url)
            repo.remotes.origin.fetch()
            repo.git.checkout(self.ref)
            repo.git.pull("origin", self.ref)
            return
        logger.debug("Cloning repository: %s", self.clone_url)
        if progress:
            clone_progress = CloneProgress()
        else:
            clone_progress = None
        no_checkout = self.ref is not None or sparse
        branch = None
        if depth is not None and self.ref is not None:
            branch = self.ref
        repo = git.Repo.clone_from(
            self.clone_url,
            dest,
            recursive=self.recursive,
            no_checkout=no_checkout,
            depth=depth,
            branch=branch,
            progress=clone_progress,
            sparse=sparse,
        )
        if self.ref:
            logger.debug("Checking out: %s", self.ref)
            repo.git.checkout(self.ref)
        if sparse:
            repo.git.sparse_checkout("init", "--cone")
            assert sparse_filter is not None
            assert isinstance(sparse_filter, list)
            repo.git.sparse_checkout("set", *sparse_filter)


class M2ISARDependency(GitDependency):
    # def __init__(self, clone_url="https://github.com/tum-ei-eda/M2-ISA-R.git", ref="coredsl_exceptions"):
    def __init__(self, clone_url="https://github.com/PhilippvK/M2-ISA-R-private.git", ref="seal5"):
        super().__init__("m2isar", clone_url, ref=ref)


DEFAULT_CDSL2LLVM_DIRS = [
    "llvm/tools/pattern-gen",
    "llvm/lib/CodeGen",
    "llvm/include/llvm/CodeGen",
    "llvm/lib/Target/RISCV",
]


def pick_coredsl2llvm_ref(ref: str, llvm_version: LLVMVersion):
    if ref is not None:
        return ref

    # auto
    if llvm_version is None:
        # TODO: warning
        return "philippvk5"

    major, minor, patch = llvm_version.triple

    if major == 20:
        ref = "llvm-20.1.0"
    elif major == 19:
        ref = "llvm-19.1.0"
    elif major == 18:
        ref = "philippvk5"

    assert ref is not None, f"Could not find a CoreDSL2LLVM ref for LLVM version {major}.{minor}.{patch}"

    return ref


class CDSL2LLVMDependency(GitDependency):
    # def __init__(self, clone_url="https://github.com/mathis-s/CoreDSL2LLVM.git", ref="main"):
    # def __init__(self, clone_url="https://github.com/PhilippvK/CoreDSL2LLVM.git", ref="philippvk5"):
    def __init__(self, clone_url="https://github.com/PhilippvK/CoreDSL2LLVM.git", ref=None, llvm_version=None):
        super().__init__("cdsl2llvm", clone_url, ref=pick_coredsl2llvm_ref(ref, llvm_version))

    def clone(
        self,
        dest: Path,
        overwrite: bool = False,
        depth: Optional[int] = None,
        progress: bool = False,
        sparse: bool = True,
        sparse_filter: Optional[List[Union[str, Path]]] = None,
    ):
        if sparse and sparse_filter is None:
            sparse_filter = DEFAULT_CDSL2LLVM_DIRS
        super().clone(
            dest, overwrite=overwrite, depth=depth, progress=progress, sparse=sparse, sparse_filter=sparse_filter
        )


# m2isar_dependency = M2ISARDependency()
# cdsl2llvm_dependency = CDSL2LLVMDependency()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Seal5"
copyright = "2024, TUM Department of Electrical and Computer Engineering - Chair of Electronic Design Automation"
author = "TUM Department of Electrical and Computer Engineering - Chair of Electronic Design Automation"
release = "v0.1.2"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "myst_parser",
    "numpydoc",
    "sphinxemoji.sphinxemoji",
]
numpydoc_show_class_members = False
autosummary_generate = True  # Turn on sphinx.ext.autosummary

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

root_doc = "index"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# html_theme = 'alabaster'
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for Exporting Seal5 artifacts"""

from seal5.logging import Logger
from seal5.wrapper import run_seal5_flow


logger = Logger("cli")


def add_wrapper_options(parser):
    """Setup parser for wrapper argument group."""
    wrapper_parser = parser.add_argument_group("wrapper options")
    wrapper_parser.add_argument(
        "files",
        nargs="*",
        help="Input files for seal5 flow",
    )
    wrapper_parser.add_argument(
        "--out-dir",
        type=str,
        default=None,
        help="Destination of artifacts",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the wrapper subcommand."""
    parser = subparsers.add_parser("wrapper", description="Run Seal5 Wrapper.")
    parser.set_defaults(func=handle)
    add_wrapper_options(parser)
    return parser


def handle(args):
    """Callback function which will be called to process the export subcommand"""
    run_seal5_flow(args.files, dest=args.dir, out_dir=args.out_dir)



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Stage times report writer for Seal5."""

import argparse
import logging
import pathlib

import pandas as pd

from seal5.settings import Seal5Settings

from seal5.logging import Logger

logger = Logger("backends.times_writer")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--fmt", type=str, choices=["auto", "csv", "pkl", "md", "mermaid"], default="auto")
    parser.add_argument("--yaml", type=str, default=None)
    parser.add_argument("--pass-times", action="store_true")
    parser.add_argument("--sum-level", type=int, default=None)
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    assert args.output is not None
    out_path = pathlib.Path(args.output)
    assert args.yaml is not None
    assert pathlib.Path(args.yaml).is_file()
    settings = Seal5Settings.from_yaml_file(args.yaml)

    def process_metrics(settings, ignore_passes: bool = False):
        metrics = settings.metrics
        assert metrics is not None
        assert isinstance(metrics, list)
        # stage_times = []
        # pass_times = []
        # for stage_metrics in metrics:
        #     assert isinstance(stage_metrics, dict)
        #     assert len(stage_metrics) == 1
        #     stage_name = list(stage_metrics.keys())[0]
        #     stage_metrics = list(stage_metrics.values())[0]
        #     assert isinstance(stage_metrics, dict)
        #     time_s = stage_metrics.get("time_s", None)
        #     start = stage_metrics.get("start", None)
        #     end = stage_metrics.get("end", None)
        #     if time_s:
        #         new = (stage_name, start, end, time_s)
        #         stage_times.append(new)
        #     if not ignore_passes:
        #         pass

        def traverse(x, prefix=None):
            assert isinstance(x, dict)
            assert len(x) == 1
            name = list(x.keys())[0]
            prefix = name if prefix is None else f"{prefix}.{name}"
            y = list(x.values())[0]
            assert isinstance(y, dict)
            passes = y.pop("passes", None)
            if passes is None:
                passes = []
            assert isinstance(passes, list)
            # ret = {}
            ret = []
            for pass_ in passes:
                ret_ = traverse(pass_, prefix=prefix)
                # ret.update(ret_)
                ret.extend(ret_)
            models = y.pop("models", None)
            if models is None:
                models = []
            assert isinstance(models, list)
            for model in models:
                ret2_ = traverse(model, prefix=prefix)
                ret.extend(ret2_)
            ret.append((prefix, y))
            return ret

        all_metrics = []
        for stage_metrics in metrics:
            result = traverse(stage_metrics)
            all_metrics.extend(result)
        filtered_metrics = [
            (x, {k: v for k, v in y.items() if k in ["time_s", "start", "end"]}) for x, y in all_metrics
        ]
        filtered_metrics = [(x, y) for x, y in filtered_metrics if len(y) > 0]
        filtered_metrics = [
            (x, {"time_s": y.get("time_s"), "start": y.get("start"), "end": y.get("end")}) for x, y in filtered_metrics
        ]
        stage_metrics = [{"stage": x, **y} for x, y in filtered_metrics if "." not in x]
        pass_metrics = [{"pass": x.split(".", 1)[-1], **y} for x, y in filtered_metrics if "." in x]
        return stage_metrics, pass_metrics

    stage_times, pass_times = process_metrics(settings, ignore_passes=not args.pass_times)

    stage_times_df = pd.DataFrame(stage_times).sort_values("start")

    if args.pass_times:
        pass_times_df = pd.DataFrame(pass_times).sort_values("start")
        if args.sum_level:
            pass_times_df["pass"] = pass_times_df["pass"].apply(lambda x: ".".join(x.split(".")[: args.sum_level]))
            pass_times_df = pass_times_df.groupby("pass", as_index=False).agg({"start": "min", "end": "max"})
            pass_times_df["time_s"] = pass_times_df["end"] - pass_times_df["start"]
            pass_times_df.sort_values("start", inplace=True)
        times_df = pd.concat([stage_times_df, pass_times_df])
    else:
        times_df = stage_times_df

    fmt = args.fmt
    if fmt == "auto":
        fmt = out_path.suffix
        assert len(fmt) > 1
        fmt = fmt[1:].lower()

    if fmt == "csv":
        times_df.to_csv(out_path, index=False)
    elif fmt == "pkl":
        times_df.to_pickle(out_path)
    elif fmt == "md":
        # times_df.to_markdown(out_path, tablefmt="grid", index=False)
        times_df.to_markdown(out_path, index=False)
    elif fmt == "mermaid":
        content = ""
        content += """gantt
  title Seal5
  dateFormat x
  axisFormat %H:%M:%S
  section Stages

"""
        cur = 0
        for _, stage_data in stage_times_df.iterrows():
            stage = stage_data.get("stage", "?")
            start = stage_data.get("start")
            end = stage_data.get("end")
            time_s = stage_data.get("time_s")
            if start:
                cur = start
            else:
                start = cur

            if end:
                cur = end
            else:
                assert time_s
                cur += time_s
                end = cur
            start = int(start * 1e3)
            end = int(end * 1e3)
            content += f"    {stage} : {start}, {end}\n"
        if args.pass_times:
            content += """
  section Passes

"""
            cur = 0
            for _, pass_data in pass_times_df.iterrows():
                pass_ = pass_data.get("pass", "?")
                start = pass_data.get("start")
                end = pass_data.get("end")
                time_s = pass_data.get("time_s")
                if start:
                    cur = start
                else:
                    start = cur

                if end:
                    cur = end
                else:
                    assert time_s
                    cur += time_s
                    end = cur
                start = int(start * 1e3)
                end = int(end * 1e3)
                content += f"    {pass_} : {start}, {end}\n"
        with open(out_path, "w") as f:
            f.write(content)

    else:
        raise ValueError(f"Unsupported fmt: {fmt}")


if __name__ == "__main__":
    main()



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utilities for handling Seal5 resources."""
from typing import Optional

import importlib_resources


# def get_patches(patch_name: str, target: Optional[str] = None, allow_missing: bool = False):
def get_patches(patch_name: Optional[str] = None, target: Optional[str] = None, allow_empty: bool = False):
    """Collect Seal5 patches."""
    ret = []
    for entry in importlib_resources.files("seal5.resources.patches").iterdir():
        # print("entry", entry.name)
        if entry.is_dir():
            if target and target != entry.name:
                continue
            for entry2 in entry.iterdir():
                # print("entry2", entry2.name)
                assert not entry2.is_dir(), "Nested patches are not allowed"
                assert entry2.is_file(), f"Invalid patch: {entry.name}"
                # if patch_name and patch_name not in entry2.stem:
                if patch_name and patch_name != entry2.stem:
                    continue
                ret.append(entry2.resolve())
        elif entry.is_file():
            ret.append(entry.resolve())

    if len(ret) == 0:
        assert allow_empty, "No patches found!"

    return ret


def get_test_cfg():
    """Get LLVM lit.cfg.py file from Seal5 resources directory."""
    ret = importlib_resources.files("seal5.resources").joinpath("lit.cfg.py")
    assert ret.is_file(), f"Test cfg not found: {ret}"
    return ret



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel by removing unused constants."""

import sys
import argparse
import logging
import pathlib

from m2isar.metamodel import patch_model

from seal5.model_utils import load_model, dump_model

from . import track_uses

from seal5.logging import Logger

logger = Logger("transform.drop_unused")


class DropUnusedContext:
    def __init__(self, names: "list[str]"):
        self.names = names
        self.to_keep = set()

    @property
    def to_drop(self):
        return set(name for name in self.names if name not in self.to_keep)

    def track(self, name: str):
        if name in self.names:
            # logger.debug("Tracked use of %s", name)
            self.to_keep.add(name)


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--compat", action="store_true")
    return parser


def run(args):
    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output

    model_obj = load_model(top_level, compat=args.compat)

    # preprocess model
    for _, set_def in model_obj.sets.items():
        logger.debug("tracking use of constants for set %s", set_def.name)
        context = DropUnusedContext(list(set_def.constants.keys()))
        patch_model(track_uses)
        for _, instr_def in set_def.instructions.items():
            logger.debug("tracking use of constants for instr %s", instr_def.name)
            instr_def.operation.generate(context)
        if len(context.to_drop) > 0:
            set_def.constants = {
                const_name: const
                for const_name, const in set_def.constants.items()
                if const_name not in context.to_drop or const_name == "XLEN"
            }
        context = DropUnusedContext(list(set_def.memories.keys()))
        for _, instr_def in set_def.instructions.items():
            logger.debug("tracking use of memories for instr %s", instr_def.name)
            instr_def.operation.generate(context)
        if len(context.to_drop) > 0:
            set_def.memories = {
                mem_name: mem for mem_name, mem in set_def.memories.items() if mem_name not in context.to_drop
            }
        context = DropUnusedContext(list(set_def.functions.keys()))
        for _, instr_def in set_def.instructions.items():
            logger.debug("tracking use of functions for instr %s", instr_def.name)
            instr_def.operation.generate(context)
        if len(context.to_drop) > 0:
            set_def.functions = {
                func_name: func for func_name, func in set_def.functions.items() if func_name not in context.to_drop
            }

    dump_model(model_obj, out_path, compat=args.compat)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import ast
from typing import Union
from pathlib import Path
import csv


def read_metrics(metrics_file: Union[str, Path], allow_missing: bool = True):
    if not Path(metrics_file).is_file():
        assert allow_missing
        return {}
    data = {}
    with open(metrics_file, "r") as infile:
        reader = csv.reader(infile)
        data = []
        for row in reader:
            data.append(row)
        assert len(data) == 2
        keys = data[0]
        vals = data[1]
        vals = list(map(lambda x: ast.literal_eval(x), vals))
        data = dict(zip(keys, vals))
    return data



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel by removing unused constants."""

import sys
import argparse
import logging
import pathlib

import pandas as pd

from m2isar.metamodel import patch_model

import seal5.model
from seal5.model_utils import load_model, dump_model

from . import visitor

from seal5.logging import Logger

logger = Logger("transform.detect_side_effects")


class VisitorContext:
    def __init__(self):
        self.is_read = False
        self.is_write = False
        self.may_load = False
        self.may_store = False
        self.is_branch = False  # TODO
        self.is_terminator = False  # TODO
        self.is_barrier = False  # TODO
        self.is_return = False  # TODO
        self.is_call = False  # TODO
        self.is_re_materializable = False
        self.is_as_cheap_as_a_move = False


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--compat", action="store_true")
    return parser


def run(args):
    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_instructions": [],
        "failed_instructions": [],
        "success_instructions": [],
    }
    for _, set_def in model_obj.sets.items():
        metrics["n_sets"] += 1
        logger.debug("collecting side effects for set %s", set_def.name)
        patch_model(visitor)
        for _, instr_def in set_def.instructions.items():
            metrics["n_instructions"] += 1
            context = VisitorContext()
            logger.debug("collecting side effects for instr %s", instr_def.name)
            try:
                instr_def.operation.generate(context)
                # TODO:
                # if arch.InstrAttribute.NO_CONT:
                # if arch.InstrAttribute.COND:
                num_ins = len(
                    [
                        op
                        for op in instr_def.operands.values()
                        if seal5.model.Seal5OperandAttribute.IN in op.attributes
                        or seal5.model.Seal5OperandAttribute.INOUT in op.attributes
                    ]
                )
                has_side_effects = num_ins == 0
                if context.may_load:
                    if seal5.model.Seal5InstrAttribute.MAY_LOAD not in instr_def.attributes:
                        instr_def.attributes[seal5.model.Seal5InstrAttribute.MAY_LOAD] = []
                if context.may_store:
                    if seal5.model.Seal5InstrAttribute.MAY_STORE not in instr_def.attributes:
                        instr_def.attributes[seal5.model.Seal5InstrAttribute.MAY_STORE] = []
                if has_side_effects:
                    if seal5.model.Seal5InstrAttribute.HAS_SIDE_EFFECTS not in instr_def.attributes:
                        instr_def.attributes[seal5.model.Seal5InstrAttribute.HAS_SIDE_EFFECTS] = []
                # TODO
                metrics["n_success"] += 1
                metrics["success_instructions"].append(instr_def.name)
            except Exception as ex:
                logger.exception(ex)
                metrics["n_failed"] += 1
                metrics["failed_instructions"].append(instr_def.name)

    dump_model(model_obj, out_path, compat=args.compat)
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""TODO"""

from m2isar.metamodel import arch, behav

# pylint: disable=unused-argument

# import sys
# import traceback


def operation(self: behav.Operation, writer):
    # print("operation", self.statements)
    if len(self.statements) > 1:
        writer.enter_block()
    for stmt in self.statements:
        stmt.generate(writer)
        if not isinstance(stmt, (behav.Conditional, behav.Operation)):
            writer.write_line(";")
    if len(self.statements) > 1:
        writer.leave_block()


def binary_operation(self: behav.BinaryOperation, writer):
    writer.write("(")
    self.left.generate(writer)
    writer.write(f") {self.op.value} (")
    self.right.generate(writer)
    writer.write(")")


def slice_operation(self: behav.SliceOperation, writer):
    # print("slice_operation")
    self.expr.generate(writer)
    writer.write("[(")
    self.left.generate(writer)
    writer.write("):(")
    self.right.generate(writer)
    writer.write(")]")


def concat_operation(self: behav.ConcatOperation, writer):
    # print("concat_operation")
    # TODO: only add () where required
    writer.write("(")
    self.left.generate(writer)
    writer.write(") :: (")
    self.right.generate(writer)
    writer.write(")")


def number_literal(self: behav.IntLiteral, writer):
    # print("number_literal")
    writer.write(self.value)


def int_literal(self: behav.IntLiteral, writer):
    # print("int_literal")
    writer.write(self.value)


def scalar_definition(self: behav.ScalarDefinition, writer):
    # print("scalar_definition", self.scalar, dir(self.scalar))
    # input("u")
    writer.write_type(self.scalar.data_type, self.scalar.size)
    writer.write(" ")
    writer.write(self.scalar.name)
    if self.scalar.value:
        writer.write(" = ")
        writer.write(self.scalar.value)
    # writer.write_line(";")


def break_(self: behav.Break, writer):
    # print("break_")
    writer.write_line("break;")


def assignment(self: behav.Assignment, writer):
    # print("assignment", self, dir(self))
    # traceback.print_stack(file=sys.stdout)
    self.target.generate(writer)
    writer.write(" = ")
    self.expr.generate(writer)
    # print("writer.text", writer.text)
    # input("assign")
    # writer.write_line(";")
    # input("123")


def conditional(self: behav.Conditional, writer):
    # print("conditional")
    for i, stmt in enumerate(self.stmts):
        if i == 0:
            writer.write("if (")
            self.conds[i].generate(writer)
            writer.write(")")
        elif 0 < i < len(self.conds):
            writer.write("else if(")
            self.conds[i].generate(writer)
            writer.write(")")
        else:
            writer.write("else")
        writer.enter_block()
        # print("stmt", stmt)
        # input("ABC")
        stmt.generate(writer)
        if not isinstance(stmt, (behav.Conditional, behav.Operation)):
            writer.write_line(";")
        nl = len(self.stmts) > i
        writer.leave_block(nl=nl)


def loop(self: behav.Loop, writer):
    # print("loop")
    writer.write("while (")
    self.cond.generate(writer)
    writer.write(")")
    writer.enter_block()
    for stmt in self.stmts:
        stmt.generate(writer)
    writer.leave_block()


def ternary(self: behav.Ternary, writer):
    # print("ternary")
    writer.write("(")
    self.cond.generate(writer)
    writer.write(") ? (")
    self.then_expr.generate(writer)
    writer.write(") : (")
    self.else_expr.generate(writer)
    writer.write(")")


def return_(self: behav.Return, writer):
    # print("return_")
    writer.write("return")
    if self.expr is not None:
        writer.write(" ")
        self.expr.generate(writer)
    # writer.write_line(";")


def unary_operation(self: behav.UnaryOperation, writer):
    # print("unary_operation")
    writer.write(self.op.value)
    writer.write("(")
    self.right.generate(writer)
    writer.write(")")


def named_reference(self: behav.NamedReference, writer):
    # print("named_reference", self.reference.name)
    writer.write(self.reference.name)
    if isinstance(self.reference, (arch.Constant, arch.Memory, arch.Scalar)):
        # writer.track(self.reference.name)
        pass
    # if isinstance(self.reference, arch.Constant):
    # 	return behav.IntLiteral(self.reference.value, self.reference.size, self.reference.signed)

    # if isinstance(self.reference, arch.Scalar) and self.reference.value is not None:
    # 		return behav.IntLiteral(self.reference.value, self.reference.size, self.reference.data_type == arch.DataType.S)


def indexed_reference(self: behav.IndexedReference, writer):
    # print("indexed_reference", self.reference.name)
    writer.write(self.reference.name)
    writer.write("[")
    # if isinstance(self.reference, arch.Memory):
    #     # writer.track(self.reference.name)
    #     pass
    self.index.generate(writer)
    writer.write("]")


def type_conv(self: behav.TypeConv, writer):
    # print("type_conv", self, dir(self))
    writer.write("(")
    writer.write_type(self.data_type, self.size)
    writer.write(")")
    writer.write("(")
    self.expr.generate(writer)
    writer.write(")")


def callable_(self: behav.Callable, writer):
    # print("callable_", self, dir(self))
    # traceback.print_stack(file=sys.stdout)
    # input("55")
    # print("args", self.args)
    # print("ron", self.ref_or_name)
    # input("q")
    ref = self.ref_or_name
    if isinstance(ref, arch.Function):
        writer.write(ref.name)
    else:
        raise NotImplementedError
    writer.write("(")
    for i, stmt in enumerate(self.args):
        stmt.generate(writer)
        if i < len(self.args) - 1:
            writer.write(", ")
    writer.write(")")


def group(self: behav.Group, writer):
    # print("group")
    # writer.enter_block()
    writer.write("(")
    self.expr.generate(writer)
    writer.write(")")
    # writer.leave_block()


def procedure_call(self: behav.ProcedureCall, context):
    # print("procedure_call")

    for arg in self.args:
        arg.generate(context)



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

from antlr4 import ParserRuleContext

from m2isar import M2DuplicateError
from .parser_gen import CoreDSL2Parser, CoreDSL2Visitor


class CoreContainerContext(ParserRuleContext):
    pass


class LoadOrder(CoreDSL2Visitor):
    def __init__(self) -> None:
        super().__init__()
        self.instruction_sets: "dict[str, CoreDSL2Parser.Instruction_setContext]" = {}
        # self.core_defs = {}
        self.stacks = {}

    def visitInstruction_set(self, ctx: CoreDSL2Parser.Instruction_setContext):
        name = ctx.name.text
        # print("set", name, [e.text for e in ctx.extension])

        if name in self.instruction_sets:
            raise M2DuplicateError(f"instruction set {name} specified more than once")

        self.instruction_sets[name] = ctx
        stack = []
        for e in ctx.extension:
            assert e.text in self.stacks, f"Set not found: {e.text}. Missing include?"
            for x in self.stacks[e.text]:
                if x not in stack:
                    stack.append(x)
            stack.append(e.text)
        self.stacks[name] = stack
        # print("stacks", self.stacks)

    # def extend_ins_set(self, ins_set_name):
    # 	if ins_set_name not in self.instruction_sets:
    # 		raise M2NameError(f"instruction set {ins_set_name} is unknown")

    # 	extensions = [e.text for e in self.instruction_sets[ins_set_name].extension]
    # 	if extensions:
    # 		ret = [ins_set_name]
    # 		for extension in extensions:
    # 			ret = self.extend_ins_set(extension) + ret
    # 		return ret
    # 	else:
    # 		return [ins_set_name]

    # def visitCore_def(self, ctx: CoreDSL2Parser.Core_defContext):
    # 	name = ctx.name.text
    # 	contributing_types = [c.text for c in ctx.contributing_types]

    # 	ins_set_queue = []
    # 	known_sets = set()

    # 	for ct in contributing_types:
    # 		new_sets = self.extend_ins_set(ct)
    # 		for new_set in new_sets:
    # 			if new_set not in known_sets:
    # 				known_sets.add(new_set)
    # 				ins_set_queue.append(self.instruction_sets[new_set])

    # 	ins_set_queue.append(ctx)
    # 	self.core_defs[name] = ins_set_queue

    def visit(self, tree):
        # print("visit", tree)
        _ = super().visit(tree)

        ret = {}

        # for core_name, contents in self.core_defs.items():
        # 	container = CoreContainerContext()
        # 	container.children = contents
        # 	container.name = core_name
        # 	ret[core_name] = container
        for set_name, set_def in self.instruction_sets.items():
            container = CoreContainerContext()
            container.children = [self.instruction_sets[x] for x in self.stacks[set_name]] + [set_def]
            container.name = set_name
            ret[set_name] = container

        return ret



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Tranform M2-ISA-R metamodel to Seal5 metamodel."""

import sys
import argparse
import logging
import pathlib

from m2isar.metamodel.code_info import CodeInfoBase
from m2isar.metamodel.utils.expr_preprocessor import process_attributes, process_functions, process_instructions

import seal5.model as seal5_model
from seal5.model_utils import load_model, dump_model

from seal5.logging import Logger

logger = Logger("transform.converter")


def convert_attrs(attrs, base):
    # print("convert_attrs", attrs)
    ret = {}
    for attr, attr_val in attrs.items():
        if isinstance(attr, str):
            attr_ = base._member_map_.get(attr.upper())
            if attr_ is not None:
                ret[attr_] = attr_val
            else:
                logger.warning("Unknown attribute: %s", attr)
                ret[attr] = attr_val
        else:
            ret[attr] = attr_val
    return ret


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--prefix", default="", type=str)
    parser.add_argument("--output", "-o", type=str, default=None)
    return parser


def run(args):
    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output

    # load model
    model_obj = load_model(top_level, compat=True)
    sets = model_obj.sets
    assert len(sets) > 0, "No sets found in model"

    # preprocess model
    for set_name, set_def in sets.items():
        logger.info("preprocessing set %s", set_name)
        process_functions(set_def)
        process_instructions(set_def)
        process_attributes(set_def)

    for set_name, set_def in sets.items():
        logger.info("replacing set %s", set_name)
        for enc, instr_def in set_def.instructions.items():
            if args.prefix:
                instr_def.name = f"{args.prefix.upper()}{instr_def.name}"
                prefix_ = args.prefix.lower().replace("_", ".")
                instr_def.mnemonic = f"{prefix_}{instr_def.mnemonic}"
            set_def.instructions[enc] = seal5_model.Seal5Instruction(
                instr_def.name,
                convert_attrs(instr_def.attributes, base=seal5_model.Seal5InstrAttribute),
                instr_def.encoding,
                instr_def.mnemonic,
                instr_def.assembly,
                instr_def.operation,
                instr_def.function_info,
                [],
                {},
            )
            set_def.instructions[enc].scalars = instr_def.scalars
        for func_name, func_def in set_def.functions.items():
            func_def.attributes = convert_attrs(func_def.attributes, base=seal5_model.Seal5FunctionAttribute)
        sets[set_name] = seal5_model.Seal5InstructionSet(
            set_def.name,
            set_def.extension,
            set_def.constants,
            set_def.memories,
            set_def.functions,
            set_def.instructions,
            {},
            {},
            {},
            {},
            {},
        )

    new_model_obj = seal5_model.Seal5Model(seal5_model.SEAL5_METAMODEL_VERSION, {}, sets, CodeInfoBase.database)
    dump_model(new_model_obj, out_path)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import inspect
import hashlib
import json
import threading
from pathlib import Path

import git
import psutil

from seal5.logging import Logger


logger = Logger("build_cache")


fuseoverlayfs = None


def init_fuseoverlayfs():
    global fuseoverlayfs
    if fuseoverlayfs is None:
        from fuseoverlayfs import FuseOverlayFS

        fuseoverlayfs = FuseOverlayFS.init()


def hash_arguments():
    args = json.dumps(
        inspect.currentframe().f_back.f_locals, sort_keys=True, default=lambda obj: f"<{obj.__class__.__name__}>"
    )
    logger.info("Build arguments: %s", args)
    return hashlib.sha1(args.encode()).hexdigest()


def get_patch_id(repo_path, base_commit, target_commit="HEAD"):
    repo = git.Repo(repo_path)
    r_fd, w_fd = os.pipe()
    t = threading.Thread(
        target=lambda: repo.git.execute(
            ["git", "diff", base_commit, target_commit], output_stream=os.fdopen(w_fd, "wb")
        )
    )
    t.start()
    # Run 'git patch-id' by piping the diff to it
    with os.fdopen(r_fd, "rb") as r:
        # patch_id will be a string like: '<patch-id> <zeroes or commit-hash>'
        patch_id = repo.git.execute(["git", "patch-id"], istream=r).split()[0]
    t.join()
    return patch_id


def combine_hashes(first, second):
    return hashlib.sha1(bytes.fromhex(first) + b":" + bytes.fromhex(second)).hexdigest()


def get_mount_info(mountpoint):
    for proc in psutil.process_iter(["name", "cmdline"]):
        try:
            # Look for fuse-overlayfs process
            if "fuse-overlayfs" in proc.info["name"]:
                cmdline = proc.info["cmdline"]
                # Check if the mountpoint is in the command line
                if any(str(mountpoint) in arg for arg in cmdline):
                    # Look for upperdir argument
                    for arg in cmdline:
                        if arg.startswith("upperdir="):
                            return Path(arg.split("=", 1)[1]).resolve().parent
        except (psutil.AccessDenied, psutil.ZombieProcess):
            continue
    return None


def get_lower_dirs(directory, default=None):
    try:
        return [Path(line.strip()) for line in (directory / "lowerdirs.txt").read_text().splitlines() if line.strip()]
    except FileNotFoundError:
        return default if default is not None else []


def save_lower_dirs(directory, lower_dirs):
    (directory / "lowerdirs.txt").write_text("\n".join(map(str, lower_dirs)))


def query_build_cache(build_hash, build_dir, cache_dir):
    init_fuseoverlayfs()

    mount_info = get_mount_info(build_dir)
    logger.debug("Mount info: %s", mount_info)
    if mount_info is not None:
        logger.debug("Mount info: %s", mount_info.name)
    logger.debug("Hash: %s", build_hash)
    overlay_dir = cache_dir / build_hash
    work_dir = cache_dir / "work"
    empty_dir = cache_dir / "empty"
    cached = False
    if mount_info is None or mount_info.name != build_hash:
        volume_dir = overlay_dir / "volume"
        if mount_info is None:
            if not overlay_dir.is_dir():
                lower_dirs = [empty_dir]
                empty_dir.mkdir(parents=True, exist_ok=True)
                volume_dir.mkdir(parents=True, exist_ok=True)
                work_dir.mkdir(parents=True, exist_ok=True)
            else:
                lower_dirs = get_lower_dirs(overlay_dir, [empty_dir])
                cached = True
        else:
            fuseoverlayfs.unmount(build_dir)
            if not overlay_dir.is_dir():
                volume_dir.mkdir(parents=True, exist_ok=True)
                lower_dirs = get_lower_dirs(mount_info)
                lower_dirs.append(mount_info / "volume")
                save_lower_dirs(overlay_dir, lower_dirs)
            else:
                lower_dirs = get_lower_dirs(overlay_dir, [empty_dir])
                cached = True
        build_dir.mkdir(parents=True, exist_ok=True)
        fuseoverlayfs.mount(build_dir, lower_dirs, workdir=work_dir, upperdir=volume_dir)
    return cached



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Filter M2-ISA-R/Seal5 metamodel."""

import re
import sys
import argparse
import logging
import pathlib

from m2isar.metamodel import arch

from seal5.model_utils import load_model, dump_model

from seal5.logging import Logger

logger = Logger("transform.filter_model")


OPCODE_LOOKUP = {
    "LOAD": 0b00000,
    "LOAD-FP": 0b00001,
    "custom-0": 0b00010,
    "MISC-MEM": 0b00011,
    "OP-IMM": 0b00100,
    "AUIPC": 0b00101,
    "OP-IMM-32": 0b00110,
    # "48bit": 0b00111,
    "STORE": 0b01000,
    "STORE-FP": 0b01001,
    "custom-1": 0b01010,
    "AMO": 0b01011,
    "OP": 0b01100,
    "LUI": 0b01101,
    "OP-32": 0b01110,
    # "64bit": 0b01111,
    "MADD": 0b10000,
    "MSUB": 0b10001,
    "NMADD": 0b10010,
    "NMSUB": 0b10011,
    "OP-FP": 0b10100,
    "OP-V": 0b10101,
    "custom-2": 0b10110,  # rv128i
    # "48bit2": 0b10111,
    "BRANCH": 0b11000,
    "JALR": 0b11001,
    # "reserved": 0b11010,
    "JAL": 0b11011,
    "SYSTEM": 0b11100,
    "OP-P": 0b11101,
    "custom-3": 0b11110,
    # "80bit+": 0b11111,
}


class DropUnusedContext:
    def __init__(self, names: "list[str]"):
        self.names = names
        self.to_keep = set()

    @property
    def to_drop(self):
        return set(name for name in self.names if name not in self.to_keep)

    def track(self, name: str):
        if name in self.names:
            # logger.debug("Tracked use of %s", name)
            self.to_keep.add(name)


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--keep-sets", type=str, default=None)
    parser.add_argument("--drop-sets", type=str, default=None)
    parser.add_argument("--keep-instructions", type=str, default=None)
    parser.add_argument("--drop-instructions", type=str, default=None)
    parser.add_argument("--keep-opcodes", type=str, default=None)
    parser.add_argument("--drop-opcodes", type=str, default=None)
    parser.add_argument("--keep-encoding-sizes", type=str, default=None)
    parser.add_argument("--drop-encoding-sizes", type=str, default=None)
    # TODO: filter builtins/aliases
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--compat", action="store_true")
    return parser


def run(args):
    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output

    model_obj = load_model(top_level, compat=args.compat)

    # preprocess model
    # print("model", model["sets"]["XCoreVMac"].keys())
    if args.keep_sets:
        keep_sets = args.keep_sets.split(",")
    else:
        keep_sets = []
    if args.drop_sets:
        drop_sets = args.drop_sets.split(",")
    else:
        drop_sets = []
    if args.keep_instructions:
        keep_instructions = args.keep_instructions.split(",")
    else:
        keep_instructions = []
    if args.drop_instructions:
        drop_instructions = args.drop_instructions.split(",")
    else:
        drop_instructions = []

    def opcodes_helper(x):
        try:
            x = int(x, 0)
        except ValueError as exc:
            x = OPCODE_LOOKUP.get(x, None)
            if x is None:
                raise ValueError(
                    f"Opcode lookup by string failed for {x}. Options: {list(OPCODE_LOOKUP.keys())}"
                ) from exc
        assert x >= 0
        assert x < 2**5
        return x

    if args.keep_opcodes:
        keep_opcodes = args.keep_opcodes.split(",")
        keep_opcodes = list(map(opcodes_helper, keep_opcodes))
    else:
        keep_opcodes = []
    if args.drop_opcodes:
        drop_opcodes = args.drop_opcodes.split(",")
        drop_opcodes = list(map(opcodes_helper, drop_opcodes))
    else:
        drop_opcodes = []
    if args.keep_encoding_sizes:
        keep_encoding_sizes = args.keep_encoding_sizes.split(",")
        keep_encoding_sizes = list(map(int, keep_encoding_sizes))
    else:
        keep_encoding_sizes = []
    if args.drop_encoding_sizes:
        drop_encoding_sizes = args.drop_encoding_sizes.split(",")
        drop_encoding_sizes = list(map(int, drop_encoding_sizes))
    else:
        drop_encoding_sizes = []
    # print("keep", keep_instructions)
    # print("drop", drop_instructions)
    # input("456")
    # def check_filter(name, keep, drop):
    #     if drop and keep:
    #         return name not in drop and name in keep
    #     elif keep:
    #         return name in keep
    #     elif drop:
    #         return name not in drop
    #     return True

    def check_filter_regex(name, keep, drop):
        if drop and keep:
            return not any(re.compile(expr).match(name) for expr in drop) and any(
                re.compile(expr).match(name) for expr in keep
            )
        if keep:
            return any(re.compile(expr).match(name) for expr in keep)
        if drop:
            return not any(re.compile(expr).match(name) for expr in drop)
        return True

    def check_encoding_filter(name, enc, keep, drop, keep2, drop2):
        opcode = None
        size = 0
        for e in reversed(enc):
            # print("e", e, dir(e))
            if isinstance(e, arch.BitVal):
                length = e.length
                if size == 0:
                    if length == 7:
                        val = e.value
                        opcode = val >> 2
            elif isinstance(e, arch.BitField):
                length = e.range.length
            else:
                assert False
            size += length
        assert size in [16, 32, 64, 128], f"Invalid size: {size} (Instruction: {name})"
        ret = True
        if drop2 and keep2:
            ret = size not in drop2 and size in keep2
        elif keep2:
            ret = size in keep2
        elif drop2:
            ret = size not in drop2
        if not ret:
            return False
        if opcode is None:  # not found (not a riscv insn?)
            return True
        if drop and keep:
            return opcode not in drop and opcode in keep
        if keep:
            return opcode in keep
        if drop:
            return opcode not in drop
        return True

    model_obj.sets = {
        set_name: set_def
        for set_name, set_def in model_obj.sets.items()
        if check_filter_regex(set_name, keep_sets, drop_sets)
    }
    for set_name, set_def in model_obj.sets.items():
        set_def.instructions = {
            key: instr_def
            for key, instr_def in set_def.instructions.items()
            if check_filter_regex(instr_def.name, keep_instructions, drop_instructions)
            and check_encoding_filter(
                instr_def.name, instr_def.encoding, keep_opcodes, drop_opcodes, keep_encoding_sizes, drop_encoding_sizes
            )
        }
        # for instr_name, instr_def in set_def.instructions.items():
    for set_name, set_def in model_obj.sets.items():
        set_def.extension = [
            extension
            for extension in set_def.extension
            if extension in model_obj.sets and len(model_obj.sets[extension].instructions) > 0
        ]

    # Remove sets without instructions
    model_obj.sets = {
        set_name: set_def
        for set_name, set_def in model_obj.sets.items()
        if len(set_def.instructions) > 0 or len(set_def.extension) > 0
    }

    dump_model(model_obj, out_path, compat=args.compat)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Console script for seal5."""

import argparse
import sys


from seal5.logging import Logger
from seal5.cli import (
    init,
    load,
    setup,
    reset,
    clean,
    patch,
    build,
    install,
    transform,
    generate,
    test,
    deploy,
    export,
    wrapper,
)
from .common import handle_logging_flags, add_common_options
from ..version import __version__

logger = Logger("cli")


def main(args=None):
    """Console script for seal5."""
    parser = argparse.ArgumentParser(
        description="Seal5 Flow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("-V", "--version", action="version", version="mlonmcu " + __version__)
    add_common_options(parser)
    subparsers = parser.add_subparsers(dest="subcommand")  # this line changed
    init.get_parser(subparsers)
    load.get_parser(subparsers)
    setup.get_parser(subparsers)
    transform.get_parser(subparsers)
    reset.get_parser(subparsers)
    clean.get_parser(subparsers)
    patch.get_parser(subparsers)
    build.get_parser(subparsers)
    install.get_parser(subparsers)
    generate.get_parser(subparsers)
    test.get_parser(subparsers)
    deploy.get_parser(subparsers)
    export.get_parser(subparsers)
    wrapper.get_parser(subparsers)
    if args:
        args = parser.parse_args(args)
    else:
        args = parser.parse_args()
    handle_logging_flags(args)

    if hasattr(args, "func"):
        args.func(args)
    else:
        logger.error("Invalid subcommand for `seal5`!")
        parser.print_help(sys.stderr)

    return 0


if __name__ == "__main__":
    sys.exit(main(args=sys.argv[1:]))  # pragma: no cover



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""PatternGen utils for seal5."""
import re
from pathlib import Path
from typing import Optional, Union
from collections import defaultdict

import yaml

from seal5.logging import Logger
from seal5.settings import PatchSettings
from seal5.types import PatchStage
from seal5.index import File, Directory, NamedPatch, write_index_yaml
from seal5.riscv_utils import build_riscv_mattr, get_riscv_defaults
from seal5 import utils

logger = Logger("tools")


def build_pattern_gen(
    src: Path,
    dest: Path,
    # debug: bool = False,
    use_ninja: bool = False,
    verbose: bool = False,
    cmake_options: Optional[dict] = None,
):
    """Build PatternGen tool."""
    if cmake_options is None:
        cmake_options = {}
    cmake_args = utils.get_cmake_args(cmake_options)
    dest.mkdir(exist_ok=True)
    utils.cmake(
        src / "llvm",
        *cmake_args,
        use_ninja=use_ninja,
        cwd=dest,
        print_func=logger.info if verbose else logger.debug,
        live=True,
    )
    utils.make(
        "pattern-gen", cwd=dest, print_func=logger.info if verbose else logger.debug, live=True, use_ninja=use_ninja
    )


def get_pattern_gen_patches(
    src: Path,
    temp_dir: Path,
    # verbose: bool = False,
):
    """Generate patches for PatternGen integration."""
    # TODO: copy!
    artifacts = []
    directory_artifact = Directory(
        "llvm/tools/pattern-gen",
        src_path=f"{src}/llvm/tools/pattern-gen",
    )
    artifacts.append(directory_artifact)
    file_artifact = File(
        "llvm/include/llvm/CodeGen/GlobalISel/PatternGen.h",
        src_path=f"{src}/llvm/include/llvm/CodeGen/GlobalISel/PatternGen.h",
    )
    artifacts.append(file_artifact)
    file_artifact = File(
        "llvm/lib/CodeGen/GlobalISel/PatternGen.cpp",
        src_path=f"{src}/llvm/lib/CodeGen/GlobalISel/PatternGen.cpp",
    )
    artifacts.append(file_artifact)
    patch_artifact = NamedPatch(
        "llvm/include/llvm/InitializePasses.h",
        key="initialize_passes_decl",
        content="void initializePatternGenPass(PassRegistry&);",
    )
    artifacts.append(patch_artifact)
    patch_artifact = NamedPatch(
        "llvm/lib/CodeGen/GlobalISel/CMakeLists.txt",
        key="gisel_cmake_srcs",
        content="  PatternGen.cpp",
    )
    artifacts.append(patch_artifact)
    patch_artifact = NamedPatch(
        "llvm/lib/CodeGen/GlobalISel/GlobalISel.cpp",
        key="gisel_init",
        content="  initializePatternGenPass(Registry);",
    )
    artifacts.append(patch_artifact)

    index_file = temp_dir / "pattern_gen_support_index.yml"
    write_index_yaml(index_file, artifacts, {}, content=True)
    patch_settings = PatchSettings(
        name="pattern_gen_support",
        stage=int(PatchStage.PHASE_0),
        comment="Integrate PatternGen in Seal5 LLVM",
        index=str(index_file),
        generated=True,
        target="llvm",
        priority=80,
    )

    return patch_settings


def build_llc(
    src: Path,
    dest: Path,
    # debug: bool = False,
    use_ninja: bool = False,
    verbose: bool = False,
    cmake_options: Optional[dict] = None,
):
    """Build llc tool."""
    if cmake_options is None:
        cmake_options = {}
    cmake_args = utils.get_cmake_args(cmake_options)
    dest.mkdir(exist_ok=True)
    utils.cmake(
        src / "llvm",
        *cmake_args,
        use_ninja=use_ninja,
        cwd=dest,
        print_func=logger.info if verbose else logger.debug,
        live=True,
    )
    utils.make("llc", cwd=dest, print_func=logger.info if verbose else logger.debug, live=True, use_ninja=use_ninja)


def run_pattern_gen(
    build_dir: Path,
    src: Path,
    dest: Union[str, Path],
    verbose: bool = False,
    ext=None,
    mattr=None,
    xlen=None,
    skip_formats=False,
    skip_patterns=False,
    skip_verify=True,
    no_extend=True,
    debug=False,
):
    """Excute pattern-gen executable."""
    if not isinstance(build_dir, Path):
        build_dir = Path(build_dir)
    pattern_gen_args = [src]
    # pattern_gen_args.extend(["-custom-legalizer-settings=foo", "-disable-gisel-legality-check"])

    if dest:
        if isinstance(dest, str):
            dest = Path(dest)
        assert dest.parent.is_dir(), f"Missing destination directory: {dest}"
        pattern_gen_args.extend(["-o", str(dest)])

    if ext:
        # pattern_gen_args.extend(["--ext", ext])
        preds = [f"Has{ext}", f"IsRV{xlen}"]
        preds_str = ", ".join(preds)
        pattern_gen_args.extend(["-p", preds_str])

    if mattr is None:
        features, _ = get_riscv_defaults()
        if ext:
            ext_ = ext.lower()
            ext_ = ext_.replace("std", "").replace("vendor", "").replace("ext", "")
            features.append(f"+{ext_}")
        mattr = build_riscv_mattr(features, xlen=xlen)

    if mattr:
        pattern_gen_args.extend(["--mattr", mattr])

    if xlen:
        assert xlen in [32, 64]
        pattern_gen_args.extend(["--riscv-xlen", str(xlen)])

    if skip_patterns:
        pattern_gen_args.append("--skip-patterns")

    if skip_formats:
        pattern_gen_args.append("--skip-formats")

    if skip_verify:
        pattern_gen_args.append("--skip-verify")

    if debug:
        pattern_gen_args.append("--debug")

    if no_extend:
        pattern_gen_args.append("--no-extend")
    pattern_gen_args.append("--stats")

    # break_on_err = True
    break_on_err = False

    # TODO: dump gmir?
    def handle_exit(code=None, out=None):
        if code is not None and code != 0:
            err_file = str(dest) + ".err"
            with open(err_file, "w", encoding="utf-8") as f:
                f.write(out)
        return code

    pattern_gen_exe = build_dir / "bin" / "pattern-gen"
    assert pattern_gen_exe.is_file(), "pattern-gen not found"

    try:
        out = utils.exec_getout(
            pattern_gen_exe,
            *pattern_gen_args,
            # cwd=dest,
            print_func=logger.info if verbose else logger.debug,
            handle_exit=handle_exit,
            live=True,
        )
    except Exception as e:
        if break_on_err:
            input("^^^ERR^^^")
        if dest.is_file():
            with open(dest, "r", encoding="utf-8") as f:
                content = f.read()
            if len(content) == 0:
                dest.unlink()
        raise e
    if not skip_patterns:
        # errs = None
        # opt_ll = None
        pat = []
        found_pattern = False
        # reason = None
        # rest = []
        is_err = False
        has_stats = False
        all_stats = defaultdict(dict)
        for line in out.split("\n"):
            # print("line", line)
            if len(line.strip()) == 0 or line.startswith("==="):
                continue
            if found_pattern:
                # print("A1")
                pat.append(line)
                # found_pattern = False
            else:
                # print("A2")
                if "Pattern for" in line:
                    # print("B1")
                    pat = [line.split(":", 1)[1]]
                    # found_pattern = True
                elif "Pattern Generation failed for" in line:
                    # reason = line
                    is_err = True
            if has_stats:
                parsed = re.compile(r"^\s*(\d+)\s([^\s]+)\s+-\s(.*)$").findall(line)
                if len(parsed) > 0:
                    assert len(parsed) == 1
                    assert len(parsed[0]) == 3
                    stat_count, stat_type, stat_descr = parsed[0]
                    stat_count = int(stat_count)
                    if stat_type not in ["pattern-gen"]:
                        continue
                    new_stats = {stat_descr: stat_count}
                    all_stats[stat_type].update(new_stats)
            else:
                if "Statistics Collected" in line:
                    has_stats = True
        if len(all_stats) > 0:
            all_stats = dict(all_stats)
            stat_file = str(dest) + ".stats"
            with open(stat_file, "w", encoding="utf-8") as f:
                yaml.dump(all_stats, f)
        if len(pat) > 0:
            pat = "\n".join(pat)
            pat_file = str(dest) + ".pat"
            with open(pat_file, "w", encoding="utf-8") as f:
                f.write(pat)

            uses = re.compile(r":\$([^:\s(),]*)").findall(pat)
            assert len(uses) > 0
            uses = list(set(uses))
            uses_str = ",".join(uses) + "\n"

            uses_file = str(dest) + ".uses"
            with open(uses_file, "w", encoding="utf-8") as f:
                f.write(uses_str)

        else:
            is_err = True
        # else:
        #     if break_on_err:
        #         print("\n".join(rest))
        #         input("^^^Pattern not found^^^")
        if is_err:
            if break_on_err:
                print(out)
                input("^^^ERROR^^^")
            dest.unlink()
        out_file = str(dest) + (".err" if is_err else ".out")
        with open(out_file, "w", encoding="utf-8") as f:
            f.write(out)
        # if len(rest) > 0:
        #     rest = "\n".join(rest)
        #     rest_file = str(dest) + (".err" if is_err else ".out")
        #     with open(rest_file, "w") as f:
        #         f.write(rest)
        # if reason:
        #     reason_file = str(dest) + ".reason"
        #     with open(reason_file, "w") as f:
        #         f.write(reason)


def convert_ll_to_gmir(
    build_dir: Path,
    src: Path,
    dest: Path,
    mattr=None,
    xlen=None,
    optimize=3,
    verbose: bool = False,
):
    """Convert LLVM-IR file to GMIR file."""
    if mattr is None:
        features, _ = get_riscv_defaults()
        mattr = build_riscv_mattr(features, xlen=xlen)

    assert xlen is not None, "Needs XLEN"

    llc_args = [src, f"-mtriple=riscv{xlen}-unknown-elf", "-stop-after=irtranslator", "-global-isel", f"-O{optimize}"]

    if mattr:
        llc_args.extend(["--mattr", mattr])

    if not isinstance(build_dir, Path):
        build_dir = Path(build_dir)

    llc_exe = build_dir / "bin" / "llc"
    assert llc_exe.is_file(), "llc  not found"

    assert dest is not None
    llc_args.extend(["-o", str(dest)])

    _ = utils.exec_getout(
        llc_exe,
        *llc_args,
        # cwd=dest,
        print_func=logger.info if verbose else logger.debug,
        live=True,
    )



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel by removing unused constants."""

from enum import IntFlag, auto


class IOMode(IntFlag):
    READ = auto()
    WRITE = auto()
